"""
Core classes and functions for tags.
"""
from .exceptions import TagError
from .signals import tag_created
from ..formats import tex_env, tex_cmd, xhtml_tag
from ..attributes import Attributes
from .utils import format_content, replace_context, copy_tag
from ..utils.string import titlelize
from ..utils.classes import weakattr, all_subclasses


class Tag(object):
    """A tag to format text in the markup document.

    Parameters
    ----------
    name : str
        The name of the tag as used in the disseminate source. (ex: 'body')
    content : Union[str, List[Union[str, list, :obj:`Tag \
        <disseminate.tags.tag.Tag>`], :obj:`Tag <disseminate.tags.tag.Tag>`]
        The contents of the tag. It can either be a string, a tag or a list
        of strings, tags and lists.
    attributes : Union[str, :obj:`Attributes \
        <disseminate.attributes.Attributes>`]
        The attributes of the tag. These are distinct from the Tag class
        attributes; they're the customization attributes specified by the user
        to change the appearance of a rendered tag. However, some or all
        attributes may be used as tag attributes, depending on the
        settings.html_valid_attributes, settings.tex_valid_attributes and so
        on.
    context : :obj:`Type[BaseContext] <.BaseContext>`
        The context with values for the document. The tag holds a weak
        reference to the context, as it doesn't own the context.

    Attributes
    ----------
    aliases : Tuple[str]
        A tuple of strings for other names a tag goes by
    hash : Optional[str]
        The hash for the tag's contents before processing. This is useful for
        detecting changes in the string's contents.
    html_name : str
        If specified, use this name for the html tag. Otherwise, use name.
    tex_cmd : str
        If specified, use this name to render the tex command.
    tex_env : str
        If specified, use this name to render the tex environment.
    tex_paragraph_newlines : bool
        If True (default), block tags within paragraphs will be typeset with
        a newline before and after. This is disable, for example, for
        equations.
    active : bool
        If True, the Tag can be used by the TagFactory.
    process_content : bool
        If True, the contents of the tag will be parsed and processed into a
        tag tree (AST) by the ProcessContent processor on creation.
    process_typography : bool
        If True, the strings in contents of the tag will be parsed and
        processed for typography characters by the ProcessTypography processor
        on creation.
    include_paragraphs : bool
        If True, then the contents of this tag can be included in paragraphs.
        See :func:`disseminate.tags.processors.process_paragraphs`.
    paragraph_role : str
        If this tag is directly within a paragraph, the type of paragraph may
        be specified with a string. The following options are possible:

        - None : The paragraph type has not been assigned
        - 'inline' : The tag is within a paragraph that includes a mix
          of strings and tags
        - 'block' : The tag is within its own paragraph.
    """

    name = None
    content = None
    attributes = None
    context = weakattr()

    aliases = None
    hash = None

    html_name = None
    tex_cmd = None
    tex_env = None
    tex_paragraph_newlines = True

    active = False

    process_macros = True
    process_content = True
    process_typography = True
    include_paragraphs = True
    paragraph_role = None

    def __init__(self, name, content, attributes, context):
        self.name = name
        self.attributes = Attributes(attributes)
        self.content = content
        self.context = context

        # Emit the tag creation signal
        tag_created.emit(tag=self, tag_base_cls=Tag, tag_factory=TagFactory)

    def __repr__(self):
        return "{type}{{{content}}}".format(type=self.name,
                                            content=self.content)

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.name == other.name and
                self.attributes == other.attributes and
                self.content == other.content and
                id(self.context) == id(other.context))

    def __getitem__(self, item):
        """Index accession."""
        if not isinstance(self.content, list):
            msg = ("Cannot access sub-tree for tag {} because "
                   "tag contents are not a list")
            raise TagError(msg.format(self.__repr__()))
        return self.content[item]

    def __len__(self):
        if isinstance(self.content, list) or isinstance(self.content, tuple):
            return len(self.content)
        else:
            return 1

    @property
    def title(self):
        """A title string converted from the content"""
        # Get the string for the content. The parent default function is called
        # to prevent recursion problems from child default methods calling
        # title.
        string = Tag.default_fmt(self)
        return titlelize(string)

    @property
    def short(self):
        """The short title for the tag"""
        short_attr = self.attributes.get('short', None)
        return short_attr if short_attr is not None else self.title

    def copy(self, new_context=None):
        """Create a copy of this tag and all sub-tabs.

        The tag copy is a deep copy of the attributes and content, but the
        context is either kept the same or replaced with the specified
        new_context.

        Parameters
        ----------
        new_context : Optional[:obj:`Type[BaseContext] <.context.BaseContext>`]
            The new context to replace, if specified.
        """
        # Copy the tag
        cp = copy_tag(tag=self)

        # Set the new context
        if new_context is not None:
            replace_context(tag=cp, new_context=new_context)
        return cp

    def flatten(self, tag=None, filter_tags=True):
        """Generate a flat list with this tag and all sub-tags and elements.

        Parameters
        ----------
        tag : Optional[:obj:`Tag <.Tag>`]
            If specified, flatten the given tag, instead of this tag.
        filter_tags : Optional[bool]
            If True, only return a list of tag objects. Otherwise, include all
            content items, including strings and lists.

        Returns
        -------
        flattened_list : List[Union[str, :obj:`Tag <.Tag>`]]
            The flattened list.
        """
        tag = tag if tag is not None else self
        flattened_list = [tag]  # add the given tag to the list

        # Process the tag's contents if present
        if hasattr(tag, 'content'):
            tag = tag.content

        # Convert the AST to a list, if it isn't already
        if not hasattr(tag, '__iter__'):
            tag = [tag]

        # Traverse the items and process each
        for item in tag:
            # Add the item to the ast
            flattened_list.append(item)

            # Process tag's sub tags or lists, if present
            if (hasattr(item, 'content') and
                (isinstance(item.content, list) or
                 isinstance(item.content, Tag))):
                flattened_list += self.flatten(tag=item.content,
                                               filter_tags=filter_tags)

        if filter_tags:
            flattened_list = [t for t in flattened_list if isinstance(t, Tag)]
        return flattened_list

    @property
    def default(self):
        return self.default_fmt()

    @property
    def txt(self):
        return self.default_fmt()

    def default_fmt(self, content=None, attributes=None):
        """Convert the tag to a text string.

        Strips the tag information and simply return the content of the tag.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.

        Returns
        -------
        text_string : str
            A text string with the tags stripped.
        """
        content = content if content is not None else self.content
        content = format_content(content=content, format_func='default_fmt')
        return ''.join(content) if isinstance(content, list) else content

    @property
    def tex(self):
        return self.tex_fmt()

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1):
        """Format the tag in LaTeX format.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.
        mathmode : Optional[bool]
            If True, the tag will be rendered in math mode. Otherwise (default)
            latex text mode is assumed.
        level : Optional[int]
            The level of the tag.

        Returns
        -------
        tex_string : str
            The formatted tex string.
        """
        # Collect the content elements
        content = content if content is not None else self.content
        content = format_content(content=content, format_func='tex_fmt',
                                 level=level + 1, mathmode=mathmode)
        content = ''.join(content) if isinstance(content, list) else content

        # Set the attributes
        attributes = self.attributes if attributes is None else attributes

        if self.tex_cmd:
            return tex_cmd(cmd=self.tex_cmd, attributes=attributes,
                           formatted_content=content)
        elif self.tex_env:
            return tex_env(env=self.tex_env, attributes=attributes,
                           formatted_content=content)
        else:
            return content

    @property
    def html(self):
        return self.html_fmt()

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1):
        """Convert the tag to an html string or html element.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.
        format_func : Optional[str]
            The format function to use with by formatted_content when
            formatting sub-tags.
        method : Optional[str]
            The rendering method for the string. ex: 'html' or 'xml'
        level : Optional[int]
            The level of the tag.

        Returns
        -------
        html : str or html element
            A string in HTML format or an HTML element (:obj:`lxml.builder.E`).
        """
        name = (self.html_name if self.html_name is not None else
                self.name.lower())

        # Collect the content elements
        content = content if content is not None else self.content
        content = format_content(content=content, format_func=format_func,
                                 level=level + 1)

        # Set the attributes
        attributes = attributes or self.attributes

        # Format the html tag
        return xhtml_tag(name=name, level=level, attributes=attributes,
                         formatted_content=content, method=method)

    @property
    def xhtml(self):
        return self.xhtml_fmt()

    def xhtml_fmt(self, method='xhtml', format_func='xhtml_fmt', **kwargs):
        """Convert the tag to an xhtml string or html element.

        Returns
        -------
        xhtml : str or xhtml element
            A string in XHTML format or an XHTML element (
            :obj:`lxml.builder.E`).
        """
        return self.html_fmt(method=method, format_func=format_func,
                             **kwargs)


class TagFactory(object):
    """Generates the appropriate tag for a given tag type.

    The tag factory instantiates tags based on loaded modules and
    initialization parameters.

    Parameters
    ----------
    tag_base_cls : :class:`Tag <disseminate.tags.tag.Tag>`
        The base class for Tag objects.
    """

    _tag_classes = None

    @classmethod
    def tag(cls, tag_name, tag_content, tag_attributes, context):
        """Return the approriate tag, given a tag_name and tag_content.

        A tag subclass, rather than the Tag base class, will be returned if

        - A tag subclass with the tag_name (or with an alias) is available.
        - The tag subclass has an 'active' attribute that is True
        - The tag's name isn't listed in the 'inactive_tags' set in the
          context.

        Parameters
        ----------
        tag_name : str
            The name of the tag. ex: 'bold', 'b'
        tag_content : Union[str, list]
            The content of the tag. ex: 'this is bold'
        tag_attributes : str
            The attributes of a tag. ex: 'width=32pt'
        context : :obj:`.document.DocumentContext`
            The document's context.

        Returns
        -------
        tag : :obj:`Tag <disseminate.tags.Tag>`
            An instance of a Tag subclass.
        """
        tag_cls = cls.tag_class(tag_name=tag_name, context=context)

        # Create the tag
        tag = tag_cls(name=tag_name, content=tag_content,
                      attributes=tag_attributes, context=context)
        return tag

    @classmethod
    def tag_class(cls, tag_name, context):
        """Retrieve the tag class for the given tag_name"""
        tag_classes = cls.tag_classes
        tag_name_lower = tag_name.lower()
        tag_cls = tag_classes().get(tag_name_lower, None)

        if (tag_cls is not None and
            getattr(tag_cls, 'active', False) and
           tag_cls.__name__.lower() not in context.get('inactive_tags', ())):
            # First, see if the tag_name matches one of the tag subclasses (or
            # tag aliases) in disseminate
            return tag_cls
        else:
            # If all else fails, just make a generic Tag.
            return Tag

    @classmethod
    def tag_classes(cls):
        """A dict of all the active Tag subclasses."""
        if cls._tag_classes is None:
            tag_classes = dict()

            for scls in all_subclasses(Tag):
                # Collect the name and aliases (alternative names) for the tag
                aliases = (list(scls.aliases) if scls.aliases is not None else
                           list())
                names = [scls.__name__.lower(), ] + aliases

                for name in names:
                    # duplicate or overwritten tag names are not allowed
                    assert name not in tag_classes
                    tag_classes[name] = scls

            cls._tag_classes = tag_classes

        return cls._tag_classes



from .composite_builder import CompositeBuilder
from ..copy import Copy


class SequentialBuilder(CompositeBuilder):
    """A composite builder that runs subbuilders in sequence (i.e. wait for
    one to finish before starting the next)

    Attributes
    ----------
    chain_on_create : bool
        If True (default), chain the parameters and outfilepath of the
        subbuilders to follow each other.
    copy : bool
        If True (default), the last subbuilders will be a Copy build to copy
        the result to the final outfilepath. This only applies if use_cache is
        False, in which case the final generated file is copied from the
        cache directory to the target directory.
    subbuilder_for_outfilename : Optional[:obj:`.builders.Builder`]
        The subbuilder instance to use for generating the outfilepath
        of the SequentialBuilder. Specifying this class name is useful for
        SequentialBuilders that use the filename hash generated by some
        subbuilders, like JinjaRender or SaveTempFile.

    Notes
    -----
    With chain_on_create is enabled, he build filepaths for subbuilders are
    set as follows, with user-supplied paths in parentheses:

    .. code-block:: text

        builder => subbuilder1 (parameters) => outfilepath1
                => subbuilder2 outfilepath2 => outfilepath3
                => subbuilder3 outfilepath3 => outfilepath4
                => outfilepath4 => (outfilepath)
    """

    action = 'sequential build'
    parallel = False

    use_cache = False
    chain_on_creation = True
    copy = True
    subbuilder_for_outfilename = None

    def __init__(self, env, **kwargs):
        super().__init__(env, **kwargs)

        # Make the last subbuilder a copy builder to copy the result of the
        # sub-builders to the final outfilepath.
        # This only applies if other subbuilders are present and the final
        # result needs to be copied to a non-cache directory--i.e. use_cache
        # is False.
        if self.copy and self.subbuilders:  # and not self.use_cache:
            cp_builder = Copy(env, use_cache=self.use_cache)
            self.subbuilders.append(cp_builder)

        # Order the subbuilders, if subbuilders are present
        if self.chain_on_creation and self.subbuilders:
            self.chain_subbuilders()

    def chain_subbuilders(self):
        """Chain the parameters and outfilepath of the subbuilders to follow
        each other"""
        # Set the parameters and outfilepaths
        current_parameters = self.parameters
        for subbuilder in self.subbuilders:
            # For the subbuilders to work together, reset their parameters
            # and outfilepath
            if current_parameters:
                subbuilder.parameters = current_parameters
            subbuilder.outfilepath = None

            # Convert the output of subbuilder into an infilepath for the
            # next subbuilder. Include the other parameters in the input
            # parameters
            outfilepath = subbuilder.outfilepath
            if outfilepath:
                current_parameters = [outfilepath] + subbuilder.not_infilepaths

        # Set the copy builder to point to the final outfilepath
        if isinstance(self.subbuilders[-1], Copy):
            self.subbuilders[-1].outfilepath = self.outfilepath

        # Chain any sequential builder subbuilders
        for subbuilder in self.subbuilders:
            if not hasattr(subbuilder, 'chain_subbuilders'):
                continue
            subbuilder.chain_subbuilders()

    @property
    def outfilepath(self):
        # Use the render builder's outfilepath if None is specified. This is
        # because the render builder's outfilepath is a hash of the input text,
        # which is unique.
        if self._outfilepath is None:
            # By default, use the parent's outfilepath, if no
            # subbuilder_for_outfilename attribute was specified
            outfilepath = CompositeBuilder.outfilepath.fget(self)

            # Otherwise use one from the subbuilders
            subbuilder = self.subbuilder_for_outfilename

            # Replace the outfilepath's filename
            if subbuilder is not None:
                sb_outfilepath = subbuilder.outfilepath
                outfilepath = outfilepath.use_name(sb_outfilepath.name)

            # Replace the extension
            out_ext = self.outfilepath_ext
            if out_ext is not None:
                outfilepath = outfilepath.use_suffix(out_ext)

            self._outfilepath = outfilepath
        return self._outfilepath

    @outfilepath.setter
    def outfilepath(self, value):
        self._outfilepath = value



"""
Tags to render equations
"""
from .img import Img
from .utils import content_to_str
from ..formats import tex_cmd, tex_env


class Eq(Img):
    r"""The inline equation tag

    Render an equation in native LaTeX (.tex targets) or into a rendered
    SVG image using LaTeX (.html targets).

    Attributes
    ----------
    aliases : Tuple[str]
        A list of strs for other names a tag goes by
    active : bool
        If True, the Tag can be used by the TagFactory.
    block_equation : bool
        If True, the equation will be rendered as a LaTeX block equation using
        a math environment. ex: \begin{equation}...\end{equation}
        If False, the equation will be rendered as a LaTeX inline equation.
        ex: "\ensuremath{y=x}
    """
    # TODO: "@eq[bold color=blue]{y=x}"

    aliases = ('term', 'termb')
    active = True
    in_ext = '.render'

    process_content = True

    bold = None
    color = None

    tex_content = None
    tex_paragraph_newlines = False

    default_block_env = "align*"

    _block_equation = False

    #: Holds a dict of sub-context dicts to be used in rendering the equation
    #: to a file. This tag owns these sub-contexts.
    _target_context = None

    def __init__(self, *args, block_equation=False, **kwargs):
        self.block_equation = block_equation
        self._target_context = dict()

        super().__init__(*args, **kwargs)

        # Convert the content into a string formatted for latex
        content = content_to_str(self.content, target='.tex',
                                 mathmode=True)
        self.content = content.strip(' \t\n')

    @property
    def block_equation(self):
        return (self._block_equation or
                'env' in self.attributes or
                self.paragraph_role == 'block')

    @block_equation.setter
    def block_equation(self, value):
        self._block_equation = value

    def tex_fmt(self, content=None, attributes=None, mathmode=False,
                **kwargs):
        # Retrieve unspecified arguments
        attributes = attributes or self.attributes
        content = content or self.content

        # Determine the environment
        env = attributes.get('env', target='.tex') or self.default_block_env

        # Add bold and color if specified
        if 'color' in attributes:
            content = tex_cmd(cmd='textcolor',
                              attributes=attributes['color'],
                              formatted_content=content)
        if 'bold' in attributes or self.name == 'termb':
            content = tex_cmd(cmd='boldsymbol', formatted_content=content)

        # Remove extra space around the content
        content = content.strip()

        if mathmode:
            return content
        else:
            if self.block_equation:
                return tex_env(env=env, attributes=attributes,
                               formatted_content=content, min_newlines=True)
            else:
                return tex_cmd(cmd='ensuremath', attributes=attributes,
                               formatted_content=content)

    def html_fmt(self, attributes=None, context=None, method='html',
                 **kwargs):
        # Create a context for the render. This context is owned by this
        # tag
        attrs = attributes or self.attributes.copy()

        # Crop equation images build. This removes white space around the
        # image so that the equation images. Note: This crop command should
        # not cut off baselines such that equation images won't line up
        # properly with the surrounding text. # ex: H vs Hy
        attrs['crop'] = (100, 100, 0, 0)
        attrs['offset'] = (0, 0, -1, 0)

        if self.block_equation:
            attrs['class'] = 'eq blockeq'
        else:
            attrs['class'] = 'eq'

        if context is None and method not in self._target_context:
            # Create a special context for this tag specifically. This tag
            # will own this context
            context = self.context.filter(['paths', 'builders'])
            context['eq'] = self
            context['template'] = 'default/tex/eq'
            self._target_context[method] = context

        context = context or self._target_context[method]
        return super(Eq, self).html_fmt(attributes=attrs, context=context,
                                        method=method, **kwargs)



"""
Terminal utilities in the CLI.
"""
import shutil
from textwrap import TextWrapper


def term_width():
    """Retrieve the current width of the terminal."""
    return shutil.get_terminal_size((80, 20))[0]  # 80-column default


def fill_string(start, end, start_len=None, end_len=None, spacer=' ',
                width=None):
    """Prints a start string and end string to fill a line on the terminal
    using the spacer character."""
    start_len = len(start) if start_len is None else start_len
    end_len = len(end) if end_len is None else end_len
    width = width if width is not None else term_width()
    return (start +
            spacer * (width - start_len - end_len) +
            end)


def wrap_with_newlines(string, wrapper=None):
    """Textwrap a string while preserving newlines."""
    wrapper = wrapper or TextWrapper(width=min(80, term_width()))
    return "\n".join([wrapper.fill(line) for line in string.splitlines()
                      if line != ""])



"""
Utilities for formatting html strings and text.
"""
from itertools import groupby

from lxml.builder import E, ElementMaker
from lxml import etree
from lxml.etree import Entity
from markupsafe import Markup

from .exceptions import FormattingError
from ..attributes import Attributes
from .. import settings


class XHtmlFormatError(FormattingError):
    """Error in xhtml formatting."""
    pass


def html_tag(*args, **kwargs):
    return html_tag(*args, **kwargs, method='html')


def xml_tag(*args, **kwargs):
    return xhtml_tag(*args, **kwargs, method='xml')


def xhtml_tag(name, attributes=None, formatted_content=None, level=1,
              target=None, method='html', nsmap=None,
              pretty_print=settings.xhtml_pretty):
    r"""Format an xhtml tag string.

    Parameters
    ----------
    name : Optional[str]
        The name of the html tag.
    attributes : Optional[Union[:obj:`Attributes <.Attributes>`, str]]
        The attributes of the tag.
    formatted_content : Optional[Union[str, list, :obj:`lxml.builder.E`]]
        The contents of the html tag.
    level : Optional[int]
        The level of the tag.
    target : Optional[str]
        If speficied, filter the attributes that match the given target.
    method : Optional[str]
        The rendering method. 'html', 'xhtml' or 'xml'
    nsmap : Optional[dict]
        Optional namespace map to create the tag with.
    pretty_print : Optional[bool]
        If True, make the formatted html pretty--i.e. with newlines and spacing
        for nested tags.

    Returns
    -------
    html : str
        If level=1, a string formatted in (x)html
        if level>1, an html element (:obj:`lxml.build.E`)

    Raises
    ------
    XHtmlFormatError : :exc:`XHtmlFormatError`
        A TagError is raised if a non-allowed environment is used

    Examples
    --------
    >>> xhtml_tag('img', attributes="src='test.svg'")
    Markup('<img src="test.svg">\n')
    >>> xhtml_tag('img', attributes="src='test.svg'", method='xml')
    Markup('<img src="test.svg"/>\n')
    """
    method = 'xml' if method == 'xhtml' else method

    # See if the tag is permitted
    allowed_tag = (name in settings.xhtml_tag_arguments or
                   name in settings.xhtml_tag_optionals)

    # Format the attributes
    attributes = '' if attributes is None else attributes
    attributes = (Attributes(attributes) if isinstance(attributes, str) else
                  attributes)

    # Get the required arguments
    if name in settings.xhtml_tag_arguments:
        # If it's an allowed tag, get the required arguments for that tag
        reqs = attributes.filter(attrs=settings.xhtml_tag_arguments[name],
                                 target=target, sort_by_attrs=True)
    elif not allowed_tag and 'span' in settings.xhtml_tag_arguments:
        # If it's not an allowed tag, use a 'span' tag and its required
        # arguments
        reqs = attributes.filter(attrs=settings.xhtml_tag_arguments['span'],
                                 target=target, sort_by_attrs=True)
    else:
        reqs = None

    # Make sure the correct number of required arguments were found
    if (reqs is not None and
       len(reqs) != len(settings.xhtml_tag_arguments[name])):
        msg = ("The html tag '{}' did not receive the correct "
               "required arguments. Required arguments received: {}")
        raise XHtmlFormatError(msg.format(name, reqs))

    # Get optional arguments
    if name in settings.xhtml_tag_optionals:
        # If it's an allowed tag, get the optional arguments for that tag
        opts = attributes.filter(attrs=settings.xhtml_tag_optionals[name],
                                 target=target, sort_by_attrs=True)
    elif not allowed_tag and 'span' in settings.xhtml_tag_optionals:
        # If it's not an allowed tag, use a 'span' tag and its optional
        # arguments
        opts = attributes.filter(attrs=settings.xhtml_tag_optionals['span'],
                                 target=target, sort_by_attrs=True)
    else:
        opts = None

    # Prepare other attributes
    other = Attributes()

    # Wrap the formatted_content in a list
    formatted_content = ([formatted_content]
                         if not isinstance(formatted_content, list) else
                         formatted_content)

    # Clean up the formatted contents
    new_formatted_content = []
    for element in formatted_content:
        # Remove empty items
        if element == '' or element is None:
            continue

        # Format safe content into an Html element
        if isinstance(element, Markup):
            element = etree.fromstring(element)

        # Append the item to the new list
        new_formatted_content.append(element)

    formatted_content.clear()
    formatted_content += new_formatted_content

    # Create the tag
    # Create the element maker with a namespace, if needed
    EM = E if nsmap is None else ElementMaker(nsmap=nsmap)

    if not allowed_tag:
        # Append the name as a class to the span element, if a class hasn't
        # been specified
        if (reqs is not None and 'class' in reqs or
           opts is not None and 'class' in opts):
            pass
        else:
            other['class'] = name

        # If the tag isn't listed in the 'allowed' tags, just create a span
        # element.
        e = EM('span', *formatted_content) if formatted_content else E('span')
    else:
        e = EM(name, *formatted_content) if formatted_content else E(name)

    # Add the reqs and opts attributes
    for attrs in (i for i in (reqs, opts, other) if i is not None):
        for k, v in attrs.items():
            if v is None:
                continue
            e.set(k, v)

    # Format the tag into a string, if it's the root level
    if level == 1:
        s = (etree.tostring(e, pretty_print=pretty_print, method=method)
                  .decode("utf-8"))
        return Markup(s)  # Mark string as safe, since it's escaped by lxml
    else:
        return e


def xml_entity(*args, **kwargs):
    return xhtml_entity(*args, **kwargs, method='xml')


def xhtml_entity(entity, level=1, method='html',
                 pretty_print=settings.xhtml_pretty):
    r"""Format an html entity string.

    Parameters
    ----------
    entity : str
        an html entity string
    level : Optional[str]
        The level of the tag.
    method : Optional[str]
        The rendering method. 'html', 'xhtml' or 'xml'
    pretty_print : Optional[bool]
        If True, make the formatted html pretty--i.e. with newlines and spacing
        for nested tags.

    Returns
    -------
    xhtml : str
        The entity formatted in xhtml.

    Raises
    ------
    XHtmlFormatError
        Raised if the contents of the tag aren't a simple string. i.e. nested
        tags are not allowed.

    Examples
    --------
    >>> xhtml_entity('alpha')
    Markup('&alpha;\n')
    """
    method = 'xml' if method == 'xhtml' else method

    if not isinstance(entity, str):
        msg = "The tag content '{}' cannot be translated into an html entity"
        raise XHtmlFormatError(msg.format(str(entity)))

    e = Entity(entity.strip())
    if level == 1:
        s = (etree.tostring(e, pretty_print=pretty_print, method=method)
             .decode("utf-8"))
        return Markup(s)  # Mark string as safe, since it's escaped by lxml
    else:
        return e


def xhtml_list(*elements, attributes=None, listtype='ol', level=1,
               target=None, method='html', pretty_print=settings.xhtml_pretty,
               inner=False):
    """
    A xhtml list element.

    Parameters
    ----------
    elements : Tuple[Tuple[int, :obj:`lxml.builder.E`]]
        Each element is a tuple of the list element level and the 'li' lxml
        element.
    attributes : Optional[Union[:obj:`Attributes <.Attributes>`, str]]
        The attributes of the tag.
    listtype : Optional[str]
        The type of list to create. ex: ul, ol
    level : Optional[int]
        The level of the tag.
    target : Optional[str]
        If speficied, filter the attributes that match the given target.
    method : Optional[str]
        The rendering method. 'html', 'xhtml' or 'xml'
    pretty_print : Optional[bool]
        If True, make the formatted html pretty--i.e. with newlines and spacing
        for nested tags.
    inner : Optional[bool]
        If True, this function is invoked as an inner html list. This is
        useful for adding html attributes only to the outer list.

    Returns
    -------
    xhtml : str
        If level=1, a string formatted in html
        if level>1, an html element (:obj:`lxml.build.E`)

    Raises
    ------
    XHtmlFormatError : :exc:`HtmlFormatError`
        A TagError is raised if a non-allowed list environment is used
    """
    method = 'xml' if method == 'xhtml' else method

    current_elements = []
    listlevel = elements[0][0] if elements else 0

    _count = 0

    # Setup a counter function for list levels. When a new list item is
    # encountered, all sub items will receive the same number so that they're
    # grouped together.
    def counter(item):
        nonlocal _count
        if item[0] == listlevel:
            _count += 1
        return _count

    # Go by each list item and its group of sublists
    for is_same_list, group in groupby(elements, counter):
        group = list(group)

        # Add the parent list item
        current_elements.append(group[0][1])

        # If there are sub list items, add these as well
        if group[1:]:
            lst = xhtml_list(*group[1:], listtype=listtype, level=level + 1,
                             target=target, method=method,
                             pretty_print=pretty_print, inner=True)
            current_elements.append(lst)

    # Wrap current_elements in a list
    attributes = attributes if not inner else ''
    if level == 1:
        e = xhtml_tag(name=listtype, formatted_content=current_elements,
                      attributes=attributes, level=level + 1, target=target,
                      method=method, pretty_print=pretty_print)
        s = (etree.tostring(e, pretty_print=pretty_print, method=method)
             .decode("utf-8"))
        return Markup(s)  # Mark string as safe, since it's escaped by lxml
    else:
        return xhtml_tag(name=listtype, formatted_content=current_elements,
                         attributes=attributes, level=level + 1, target=target,
                         method=method, pretty_print=pretty_print)



from . import target_builder
from . import signals, receivers
from .html_builder import HtmlBuilder
from .xhtml_builder import XHtmlBuilder
from .epub_builder import EpubBuilder
from .tex_builder import TexBuilder
from .pdf_builder import PdfBuilder
from .txt_builder import TxtBuilder

__all__ = ('target_builder', 'signals', 'receivers', 'HtmlBuilder',
           'XHtmlBuilder', 'EpubBuilder', 'TexBuilder', 'PdfBuilder',
           'TxtBuilder')



"""
A handler to pygmentize a source page
"""
from pathlib import Path

from pygments import highlight
from pygments.lexers import get_lexer_for_filename
from pygments.formatters import HtmlFormatter

from .server import ServerHandler


class PygmentizeHandler(ServerHandler):
    """A handler to pygmentize source files"""

    def get(self, path):
        # Get the appropriate lexer
        lexer = get_lexer_for_filename(path, stripall=True)

        # Setup the formatter
        formatter = HtmlFormatter(linenos=True, cssclass="source", full=True)

        # Source file
        code = Path(path).read_text()

        self.write(highlight(code, lexer, formatter))



"""
Utilities for dealing with Python classes.
"""
import weakref
from itertools import chain as i_chain


def all_subclasses(cls):
    """Retrieve all subclasses, sub-subclasses and so on for a class

    Parameters
    ----------
    cls : Type
        The class object to inspect for subclasses.

    Returns
    -------
    subclasses : list
        The list of all subclasses.

    Examples
    --------
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> all_subclasses(A)
    [<class 'disseminate.utils.classes.B'>, \
<class 'disseminate.utils.classes.C'>]
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__()
                                   for g in all_subclasses(s)]


def all_parent_classes(cls):
    """Retrieve a list of all of parent classes for a given class.

    Parameters
    ----------
    cls : Type
        The class object to inspect for parent classes.

    Returns
    -------
    parent_classes : list
        The list of all parent classes (not including the object class).

    Examples
    --------
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> all_parent_classes(C)
    [<class 'disseminate.utils.classes.B'>, \
<class 'disseminate.utils.classes.A'>]
    """
    return list(cls.__bases__) + [g for s in list(cls.__bases__)
                                  for g in all_parent_classes(s)
                                  if g != object]


def all_attributes_values(cls, attribute):
    """Retrieve a list of all values for a given class's attribute
    for the given class and all its parent classes.

    Parameters
    ----------
    cls : Type
        The class object to inspect for parent classes.
    attribute : str
        The attribute whose value should be retrieved from each class.

    Returns
    -------
    parent_attributes : list
        The list of all parent classe attributes.

    Examples
    --------
    >>> class A(object):
    ...     value = 'a'
    >>> class B(A):
    ...     value = 'b'
    ...     value2 = (1, 2)
    >>> class C(B):
    ...     value = 'c'
    ...     value2 = (2, 3)
    >>> sorted(all_attributes_values(C, 'value'))
    ['a', 'b', 'c']
    >>> sorted(all_attributes_values(C, 'value2'))
    [1, 2, 3]
    """
    bases = [cls] + all_parent_classes(cls)
    return set(i_chain(*[getattr(b, attribute) for b in bases
                         if hasattr(b, attribute)]))


def all_dicts(cls):
    """Produce a dict with entries from a given class's __dict__ and the
    __dict__ of all its parent classes.

    Precedence is given to the subclasses over parent classes if the attributes
    were overwritten by subclasses.

    Examples
    --------
    >>> class A(object):
    ...     a1 = 1
    ...     a2 = 2
    >>> class B(A):
    ...     a2 = 3
    ...     b1 = 4
    >>> d = all_dicts(B)
    >>> d['a1']
    1
    >>> d['a2']
    3
    >>> d['b1']
    4
    """
    bases = all_parent_classes(cls)[::-1] + [cls]
    d = dict()
    for base in bases:
        if hasattr(base, '__dict__'):
            d.update(base.__dict__)
    return d


class weakattr(object):
    """A descriptor to store a weakref to an object attribute"""

    def __get__(self, obj, objtype=None):
        weakref_dict = obj.__dict__.setdefault('__weakrefattrs__', dict())
        attrname = self.attrname(obj)
        value = weakref_dict.get(attrname, None)
        return value() if callable(value) else None

    def __set__(self, obj, value):
        if value is None:  # do nothing is a value of None is assigned
            return
        weakref_dict = obj.__dict__.setdefault('__weakrefattrs__', dict())
        attrname = self.attrname(obj)
        weakref_dict[attrname] = weakref.ref(value)

    def __delete__(self, obj):
        weakref_dict = obj.__dict__.setdefault('__weakrefattrs__', dict())
        attrname = self.attrname(obj)
        del weakref_dict[attrname]

    def attrname(self, obj):
        if not hasattr(self, '_attrname'):
            cls = obj.__class__
            for name, value in all_dicts(cls).items():
                if value == self:
                    self._attrname = name
        if hasattr(self, '_attrname'):
            return self._attrname
        else:
            raise AttributeError("weakrefattr could not be found in "
                                 "object '{}'".format(obj))



"""
A scanner for html files.
"""
import regex

from .scanner import Scanner


#: regex for processing <link> tags in html headers
#: matches: '<link rel="stylesheet" href="/media/css/bootstrap.min.css">'
#: matches: '<link href="{{ "/media/css/default.css" | rewrite_path }}">'
_re_html_link = regex.compile(r'<link\s+(?:[^>]*?\s+)?'
                              r'href=\s*[\"\'\s\{]+'
                              r'(.*?)'
                              r'[\"\']')

#: regex for processing <img> tags in html templates
#: matches: '<img src="/media/icons/menu_inactive.svg">'
#: matches: '<img src="{{ "/media/icons/menu_inactive.svg" | rewrite_path }}">'
_re_html_img = regex.compile(r'<img\s+(?:[^>]*?\s+)?'
                             r'src=\s*[\"\'\s\{]+'
                             r'(.*?)'
                             r'[\"\']')


class HtmlScanner(Scanner):
    """A scanner for html files."""

    extensions = ('.html', '.htm', '.xhtml')

    @staticmethod
    def scan_function(content):
        """Scan additional file dependencies from the content of an html
        file."""
        filepaths = []

        # Find the matches
        for re_func in (_re_html_link, _re_html_img):
            for match in re_func.finditer(content):
                # Retrieve the link
                link = match.groups()[0]

                # Skip protocols; we're looking for actual files
                if '://' in link:
                    continue

                filepaths.append(link)
        return filepaths



"""
Common file options for CLI sub-commands.
"""
import click

from ..utils.string import nicejoin

debug_option = click.option('--debug', default=False, is_flag=True,
                            help="Show debugging information")


# Project path options


in_option = click.option('--in-path', '-i', required=False,
                         type=click.Path(exists=True, file_okay=True,
                                         dir_okay=True),
                         default='.',
                         help="the directory or file path for a project root "
                              "document")
out_option = click.option('--out-dir', '-o', required=False,
                          type=click.Path(exists=True, file_okay=False,
                                          dir_okay=True),
                          default=None,
                          help="the target directory for the generated output "
                               "documents")

_file_options = [in_option, out_option]


def file_options(func):
    """Wrap the file options in a decorator"""
    for option in reversed(_file_options):
        func = option(func)
    return func


def check_out_dir(root_docs, out_dir):
    """Check that there is only 1 project when specifying the output directory.

    Multiple projects cannot (currently) write to the same output directory.
    """
    if len(root_docs) > 1 and out_dir is not None:
        docs_string = [repr(d) for d in root_docs]
        msg = ("The '-o' flag only work when one project is specified. "
               "The following projects are currently "
               "loaded: {}".format(nicejoin(*docs_string)))
        raise click.BadParameter(msg)



import pathlib

from .utils.types import IntPositionalValue, StringPositionalValue
from .__version__ import __version__


#: Document Defaults
#: -----------------

#: The default extension of markup files
document_extension = '.dm'

#: The default directory for source files
document_src_directory = 'src'

#: Maximum file size
document_max_size = 204800  # 200kB

#: If True, directories for target files will be created, if they not already
#: exist
create_dirs = True

#: HTTP Server
#: -----------

#: default port on the localhost to listen for http requests
default_port = 8899

#: A list of extensions that will be sent with the 'text/plain' MIME type
text_extensions = ['.tex', ]

#: CLI
#: ---

#: Use terminal colors for the CLI
colored_term = True

#: Options for the 'init's subcommand
cli_init_starter_name_color = 'red'
cli_init_starter_subheadind_color = 'magenta'

#: Convert and Tag Defaults
#: ------------------------

#: The default prefix for tags
tag_prefix = r'@'

#: Maximum depth of tag trees
tag_max_depth = 30

# If True, converted files will be updated only when they're changed. Otherwise
# converted files will always be updated
convert_cache = True

#: The location in the target_root to store temporary cached files
cache_path = '.cache'

#: The default location for media (image, css, js) files.
media_path = 'media'


#: Attributes
#: ----------

#: For attribute entries that are target-specific, use the following character
#: or string to separate the key and target. ex: class.html
attribute_target_sep = '.'

#: Context
#: -------

#: Attribute in the context in which the body of a file (the string and AST)
#: is stored in.
body_attr = 'body'

#: Maximum depth of loaded context dicts
context_max_depth = 6

#: Maximum file size for context header files
context_max_size = 8192  # 8kB

default_context = {
    # Set the disseminate version
    'version': __version__,

    # The default targets to render. This is a string so that documents
    # can overwrite these values
    'targets': {'html'},

    'paths': [],
    'template': 'default',

    # Entries for navigation
    'prev': '',
    'next': '',
    'srclink': '',
    'txtlink': '',
    'texlink': '',
    'pdflink': '',
    'epublink': '',

    'toc': 'all headings expanded',

    # Options related to links
    'relative_links': True,
    'base_url': '/{target}/{subpath}',
    'media_path': 'media',

    # Process tags for the following entries in a context
    # (see processors/process_context_tags.py)
    'process_context_tags': {body_attr, 'toc', 'prev', 'next',
                             'srclink', 'txtlink', 'texlink', 'pdflink',
                             'epublink'},

    # Process paragraphs for tags with the following names
    # (see tags/paragraphs.py)
    'process_paragraphs': {body_attr, "featurebox"},

    # The following are strings to present labels. They are substituted with
    # values from their respective label and parsed in disseminate format.
    # The label format strings are identified first by the tag that uses
    # the label, then the kind(s) of the label, and finally, and optionally,
    # the target for the rendered document.
    'label_fmts': {
        'document': '@label.title',

        'ref_document': '@label.title',

        # caption tags
        'caption_figure': '@b{Fig. @label.number}. ',
        'caption_table': '@b{Table @label.number}. ',

        'ref_caption_figure': '@b{Fig. @label.number}',
        'ref_caption_table': '@b{Table @label.number}',

        # Headings tags
        'heading': '@label.tree_number. @label.title',
        'heading_title': '@label.title',
        'heading_part': 'Part @label.part_number. @label.title',
        'heading_chapter': 'Chapter @label.chapter_number. @label.title',

        'ref_heading': '@label.title',
        'ref_heading_part': 'Part @label.part_number. @label.title',
        'ref_heading_chapter': 'Chapter @label.chapter_number. @label.title',

        # Citation formats
        # 'citation': '@sup{label.number}',
        #
        # 'citation_ref': '({refs})',
        # 'citation_id': 'citation.number',
        # 'citation_ref_sep': ',',
        #
        # 'citation_journal': '$journal.authors, $journal.title, '
        #                     '@b{$journal.journal}',
        #
        # 'citation_author_sep': ',',
        # 'citation_author_and': True,
        # 'citation_author_initials_point': True,

    },

    'label_resets': {
        # New chapters reset the numbers on sections, subsections,
        # subsubsections
        'part': {'chapter', 'section', 'subsection', 'subsubsection'},
        'chapter': {'section', 'subsection', 'subsubsection', 'figure',
                    'table'},
        'section': {'subsection', 'subsubsection'},
        'subsection': {'subsubsection'},
    },

    # Set the separator between a doc_id and label_id when identifying a label
    'label_sep': '::',

    # The following tags are unavailable. This is a string so that contexts
    # from templates can replace these values. See
    # disseminate.tags.factory.TagFactory.
    'inactive_tags': set(),

    # Target-specific customizations
    'epub': dict(),

    # Macros are string entries that aren't processed into tags and asts.
    # These start with the 'tag_prefix' (e.g. '@')
    # Macros - Isotopes
    '@1H': '@sup{1}H', '@2H': '@sup{2}H', '@13C': '@sup{13}C',
    '@15N': '@sup{15}N', '@19F': '@sup{19}F', '@31P': '@sup{31}P',

    # Macros - Molecules
    '@H2O': 'H@sub{2}O',

    # Macros - Symbols
    '@deg': '@sup{○}',
    '@degC': '@sup{○}C',
    '@times': '@symbol{times}',

    # Macros - Greek
    '@alpha': '@symbol{alpha}', '@beta': '@symbol{beta}',
    '@gamma': '@symbol{gamma}', '@delta': '@symbol{delta}',
    '@epsilon': '@symbol{epsilon}', '@zeta': '@symbol{zeta}',
    '@eta': '@symbol{eta}', '@theta': '@symbol{theta}',
    '@iota': '@symbol{iota}', '@kappa': '@symbol{kappa}',
    '@lambda': '@symbol{lambda}', '@mu': '@symbol{mu}', '@nu': '@symbol{nu}',
    '@xi': '@symbol{xi}', '@omicron': '@symbol{omicron}', '@pi': '@symbol{pi}',
    '@rho': '@symbol{rho}', '@sigma': '@symbol{sigma}', '@tau': '@symbol{tau}',
    '@upsilon': '@symbol{upsilon}', '@phi': '@symbol{phi}',
    '@chi': '@symbol{chi}', '@psi': '@symbol{psi}', '@omega': '@symbol{omega}',
    '@Alpha': '@symbol{Alpha}', '@Beta': '@symbol{Beta}',
    '@Gamma': '@symbol{Gamma}', '@Delta': '@symbol{Delta}',
    '@Epsilon': '@symbol{Epsilon}', '@Zeta': '@symbol{Zeta}',
    '@Eta': '@symbol{Eta}', '@Theta': '@symbol{Theta}',
    '@Iota': '@symbol{Iota}', '@Kappa': '@symbol{Kappa}',
    '@Lambda': '@symbol{Lambda}', '@Mu': '@symbol{Mu}', '@Nu': '@symbol{Nu}',
    '@Xi': '@symbol{Xi}', '@Omicron': '@symbol{Omicron}', '@Pi': '@symbol{Pi}',
    '@Rho': '@symbol{Rho}', '@Sigma': '@symbol{Sigma}', '@Tau': '@symbol{Tau}',
    '@Upsilon': '@symbol{Upsilon}', '@Phi': '@symbol{Phi}',
    '@Chi': '@symbol{Chi}', '@Psi': '@symbol{Psi}', '@Omega': '@symbol{Omega}',

}

#: Template Defaults
#: -----------------

#: default basename for a template file
template_basename = 'template'

#: default filename for additional context files from templates
template_context_filename = 'context.txt'

#: default template starter description file
template_starter_dir = 'starters'
template_starter_desc_filename = 'description.yaml'

#: Use module templates only. If True, user-specified custom templates, which
#: may not be safe, can be used.
module_only = False

#: the paths for templates in disseminate modules
module_template_paths = [pathlib.Path(__file__).parent / 'templates']

#: Builder Defaults
#: -------------------

#: A series of allowed tracked extensions for each target type with information
#: to translate these files to a useable form, if needed.
#: The keys are the targets. The values are lists of valid extensions that
#: can be included for the target, in order of decreasing preference.
tracked_deps = {
    # html targets ca use .css style files, .svg and .png images
    '.html': ['.css', '.svg', '.png'],
    '.xhtml': ['.css', '.svg', '.png'],
    # tex (and pdf) target can use .pdf and .png images
    '.tex': ['.pdf', '.png'],
    # css files can include .css files
    '.css': ['.css', ]}

#: The default decider class
default_decider = 'Md5Decider'

#: The default number of seconds before a subprocess is timedout.
default_timeout = 15

#: Tags
#: ----

empty = tuple()

#: The number of spaces to identify sublist items.
list_level_spaces = 2

#: XHTML targets
#: ~~~~~~~~~~~~

xhtml_namespace = {'epub': 'http://www.idpf.org/2007/ops'}

#: Render XHTML pages with newlines and indentation
xhtml_pretty = True

#: Allowed xhtml tags with required arguments/attributes.
#: This dict will be checked to see if an html tag is allowed.
#: The values are tuples that indicate the order of attributes for tags
xhtml_tag_arguments = {'a': ('href',),
                       'img': ('src',),
                       'link': ('rel',)
                       }

#: Allowed optional arguments/attributes for html tags
#: This dict will be checked to see if an html tag is allowed.
epub_type = "{{{}}}type".format(xhtml_namespace['epub'])
xhtml_tag_optionals = {'a': ('class', 'role', epub_type),
                       'aside': ('class', 'id', epub_type),
                       'blockquote': empty,
                       'br': empty,
                       'code': empty,
                       'caption': ('class', 'id'),
                       'dd': empty,
                       'div': ('class', 'id'),
                       'dl': empty,
                       'dt': empty,
                       'em': empty,
                       'figure': ('id', 'class'),
                       'figcaption': ('id', 'class'),
                       'h1': ('id', 'class'),
                       'h2': ('id', 'class'),
                       'h3': ('id', 'class'),
                       'h4': ('id', 'class'),
                       'h5': ('id', 'class'),
                       'hr': empty,
                       'li': ('class',),
                       'link': ('href', 'media'),
                       'i': empty,
                       'img': ('alt', 'class', 'style'),
                       'ol': ('class',),
                       'p': ('class',),
                       'pre': ('class',),
                       'span': ('class', 'id', 'style'),
                       'strong': empty,
                       'sub': empty,
                       'sup': empty,
                       'table': ('id', 'class',),
                       'tbody': ('class',),
                       'td': ('class',),
                       'th': ('class',),
                       'thead': ('class',),
                       'tr': ('class',),
                       'ul': ('class',),
                       }

#: TEX targets
#: ~~~~~~~~~~~

#: Allowed latex environments and required arguments. If an environment is not
#: listed here or in the tex_env_optionals, it cannot be used.
#: The values are tuples that indicate the order of attributes for environments
#: and commands
tex_env_arguments = {'align': empty,  # no required arguments
                     'align*': empty,
                     'alignat': (IntPositionalValue,),  # \begin{alignat}{3}
                     'alignat*': (IntPositionalValue,),  # \begin{alignat*}{3}
                     'center': empty,
                     'enumerate': empty,

                     'examplebox': empty,
                     'featurebox': empty,
                     'problembox': empty,

                     'itemize': empty,
                     'panel': (StringPositionalValue,),
                     'tabular': (StringPositionalValue,),
                     'table': empty,
                     'table*': empty,
                     'margintable': empty,
                     'toclist': empty,
                     'verbatim': empty,
                     }

tex_env_optionals = {'enumerate': (StringPositionalValue, 'label'),
                     'easylist': (StringPositionalValue,),
                     'figure': (StringPositionalValue,),  # \begin{figure}[h]
                     'figure*': (StringPositionalValue,),
                     'marginfigure': (StringPositionalValue,),
                     }

tex_cmd_arguments = {'textbf': empty,
                     'textit': empty,

                     'part': empty,
                     'chapter': empty,
                     'section': empty,
                     'subsection': empty,
                     'subsubsection': empty,
                     'paragraph': empty,
                     'subparagraph': empty,

                     'part*': empty,
                     'chapter*': empty,
                     'section*': empty,
                     'subsection*': empty,
                     'subsubsection*': empty,

                     'maketitle': empty,
                     'title': empty,
                     'author': empty,
                     'today': empty,

                     'caption': empty,
                     'ensuremath': empty,
                     'marginnote': empty,
                     'marginpar': empty,

                     'item': empty,
                     'setcounter': (StringPositionalValue, IntPositionalValue),
                     'label': empty,
                     'includegraphics': empty,
                     'textcolor': (StringPositionalValue,),
                     'boldsymbol': empty,

                     'toprule': empty,
                     'midrule': empty,
                     'bottomrule': empty,

                     'hfill': empty,
                     'pageref': (StringPositionalValue,),
                     'href': (StringPositionalValue,),

                     'detokenize': empty,
                     }

tex_cmd_optionals = {'includegraphics': ('width', 'height'),
                     }


#: Default text width to rewrap paragraphs. Set to 0 to disable.
# tex_paragraph_width = 80



"""
Command-line interface tools for Signals
"""
from textwrap import TextWrapper
from string import ascii_lowercase

from click import style

from ..term import term_width
from ...utils.string import stub


def print_signals(signal_namespace):
    """Given an signal namespace, print all signals and connected receivers.

    Parameters
    ----------
    signal_namespace : :obj:`.signals.signals.Namespace`
        The namespace for all signals.

    Returns
    -------
    None
    """

    wrap_desc = TextWrapper(initial_indent=' ' * 4,
                            subsequent_indent=' ' * 4,
                            width=term_width())
    wrap_fields = TextWrapper(initial_indent=' ' * 4,
                              subsequent_indent=' ' * 7,
                              width=term_width())
    lowercase = list(ascii_lowercase)

    for count, (name, signal) in enumerate(signal_namespace.items(), 1):
        msg = "{count}. ".format(count=count)
        msg += style(name, bold=True, underline=True)
        msg += '\n'

        # Get the docstring (first line), if available
        if getattr(signal, '__doc__', None) is not None:
            docstring = stub(signal.__doc__)
            msg += wrap_desc.fill(docstring) + '\n'

        # Get the receivers for the signal
        receivers = list(enumerate(sorted(signal.receivers_dict().items())))

        if len(receivers) > 0:
            rec_str = style("Receivers:", bold=True)
            msg += "\n" + wrap_desc.fill(rec_str) + "\n"

        for count, (order, receiver) in receivers:
            rec_str = (lowercase[count] if count < len(lowercase) else
                       str(count))
            rec_str += ". "
            rec_str += style("{}".format(receiver.__name__), fg='cyan')

            if receiver.__doc__ is not None:
                docstring = stub(receiver.__doc__)
                rec_str += " - " + docstring
            msg += wrap_fields.fill(rec_str) + "\n"

            rec_str = "   order: " + str(order)
            msg += wrap_fields.fill(rec_str) + '\n'

        print(msg)



"""
A builder for document targets.
"""
import logging

from ..jinja_render import JinjaRender
from ..composite_builders import SequentialBuilder, ParallelBuilder
from ..utils import generate_outfilepath
from ...utils.classes import weakattr


class TargetBuilder(SequentialBuilder):
    """A builder for a document target, like html, tex or pdf

    Parameters
    ----------
    env: :obj:`.builders.Environment`
        The build environment
    context: :obj:`.context.Context`
        The context dict for the document being rendered.
    parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including filepaths, for the build
    outfilepath : Optional[:obj:`pathlib.Path`]
        If specified, the path for the output file.
    subbuilders : List[:obj:`.builders.Builder`]
        A list of subbuilders to add to this TargetBuilder.

    Attributes
    ----------
    only_root : Optional[bool]
        If True, only add the target builder for the root document.
    add_parallel_builder : Optional[bool]
        If True (default), create a parallel builder for adding dependencies
        for a target.
    add_render_builder : Optional[bool]
        If True (default), create a render builder to render the target.
    """

    active_requirements = ('priority',)
    priority = 1000
    context = weakattr()

    use_media = False
    use_cache = False
    chain_on_creation = False
    copy = False
    clear_done = False

    only_root = False
    add_parallel_builder = True
    add_render_builder = True

    parameters_from_signals = 'ref_label_dependencies'

    _parallel_builder = None
    _render_builder = None

    def __init__(self, env, context, parameters=None, outfilepath=None,
                 subbuilders=None, use_cache=None, **kwargs):
        if __debug__ and 'src_filepath' in context:
            # log the creation of the target builder
            msg = "Add target builder '{}' to document '{}'"
            logging.debug(msg.format(self.__class__.__name__,
                                     context['src_filepath']))

        # Configure the parameters
        self.context = context
        parameters = parameters if parameters is not None else []
        parameters = (list(parameters) if isinstance(parameters, tuple) or
                      isinstance(parameters, list) else [parameters])
        subbuilders = subbuilders or []

        # Determine the target for the TargetBuilder
        if 'target' in kwargs:
            pass
        elif 'target' not in kwargs and self.target is not None:
            kwargs['target'] = self.target
        elif 'target' not in kwargs and self.outfilepath_ext is not None:
            kwargs['target'] = self.outfilepath_ext
        else:
            msg = ("A target or outfilepath_ext attribute must be specified "
                   "for a TargetBuilder.")
            raise AssertionError(msg)

        # Add the target_builder to the context
        builders = context.setdefault('builders', dict())
        builders[self.outfilepath_ext] = self

        # Setup parameters specific to this target builder.
        # Make sure the src_filepath is in the parameters, since if this
        # changes, it changes the input parameter hash for the decider
        if len(parameters) == 0 and 'src_filepath' in context:
            parameters.append(context['src_filepath'])

        # Make sure the name of the builder is included in the parameters,
        # since this will change the build hash for different target builders
        parameters.insert(0, "build '{}'".format(self.__class__.__name__))

        # Determine if this target builder matches a document target
        is_document_target = self.outfilepath_ext in context.targets

        # Only use cache for outfilepaths if this is not a document target
        if use_cache is None:
            use_cache = False if is_document_target else True

        # Setup the outfilepath, if one isn't specified.
        if outfilepath is None:
            outfilepath = generate_outfilepath(env=env,
                                               parameters=parameters,
                                               target=self.outfilepath_ext,
                                               ext=self.outfilepath_ext,
                                               use_cache=use_cache,
                                               use_media=self.use_media)

        # Add a parallel builder for dependencies
        if self.add_parallel_builder:
            parallel_builder = ParallelBuilder(env=env, **kwargs)
            subbuilders.append(parallel_builder)
            self._parallel_builder = parallel_builder

        # Add a render builder for the final jinja file
        if self.add_render_builder:
            render_builder = JinjaRender(env, context, outfilepath=outfilepath,
                                         render_ext=self.outfilepath_ext,
                                         **kwargs)
            # Add ref label dependencies
            render_builder.parameters_from_signals = 'ref_label_dependencies'

            subbuilders.append(render_builder)
            self._render_builder = render_builder

        # Initialize builder
        super().__init__(env, parameters=parameters, outfilepath=outfilepath,
                         subbuilders=subbuilders, use_cache=use_cache,
                         **kwargs)

    @property
    def outfilepath(self):
        if self._render_builder is not None:
            return self._render_builder.outfilepath
        else:
            return SequentialBuilder.outfilepath.fget(self)

    @outfilepath.setter
    def outfilepath(self, value):
        if self._render_builder is not None:
            self._render_builder.outfilepath = value
        else:
            SequentialBuilder.outfilepath.fset(self, value)

    def add_build(self, parameters, outfilepath=None, context=None,
                  use_cache=None, **kwargs):
        """Create and add a sub-builder to the composite builder."""
        use_cache = use_cache or self.use_cache

        par_builder = self._parallel_builder
        builder = par_builder.add_build(parameters=parameters,
                                        outfilepath=outfilepath,
                                        context=context,
                                        use_cache=use_cache,
                                        **kwargs)
        return builder

    def build(self, complete=False):
        # Reload the document
        context = self.context
        document = getattr(context, 'document', None)
        if document is not None:
            document.load()

        # Run the build
        status = super().build(complete=complete)

        return status

    def chain_subbuilders(self):
        return None



"""
A decider that uses MD5 hashes.
"""
import diskcache

from .decider import Decider, Decision
from .utils_hash import hash_items


class Md5Decision(Decision):
    """A decision for the Md5Decider"""

    def build_needed(self, inputs, output, reset=False):
        build_needed = super().build_needed(inputs, output)

        if build_needed:
            # A build is needed if some of the files don't exist
            return True

        # At this stage, the files exist but we're not sure whether their
        # md5 hashes have changed. See if there's an existing (cached) hash
        db = self.parent_decider.db
        assert db is not None

        # Check to see there's an existing hash
        input_hash, output_hash = self.calculate_hash(inputs=inputs,
                                                      output=output)

        # Check the database hash (input_hash is the key, output_hash is the
        # value)
        key = input_hash
        cached_output_hash = db.get(key, None)

        # Reset the cached hash, if needed
        if reset:
            # Recalculate the hash
            input_hash, output_hash = self.calculate_hash(inputs=inputs,
                                                          output=output)
            db[input_hash] = output_hash
            return False
        else:
            return cached_output_hash != output_hash

    @staticmethod
    def calculate_hash(inputs, output):
        """Calculate the md5 hash for the inputs, output and args."""
        sorted_inputs = sorted(inputs, key=str)
        hash_input = hash_items(*sorted_inputs)
        hash_output = hash_items(output)

        return (hash_input, hash_output)


class Md5Decider(Decider):
    """A decider that uses md5 hashes to compute whether a build is needed."""

    name = None
    decision_cls = Md5Decision

    _db = None

    def __init__(self, env, name=None):
        self.name = name
        super().__init__(env)

    def __del__(self):
        # Close the database if it's open
        if self._db is not None:
            self.db.close()

    @property
    def db_path(self):
        """The path for the database file"""
        name = self.name if self.name is not None else 'md5decider'
        return str(self.environment.cache_path / name)

    @property
    def db(self):
        if self._db is None:
            self._db = diskcache.Cache(self.db_path)
        return self._db



"""
Receivers for processing a tag's macros on tag creation.
"""
from ..signals import tag_created
from ...utils.string import replace_macros


@tag_created.connect_via(order=100)
def process_macros(tag, **kwargs):
    """A receiver for replacing macros in pre-parsed tag strings."""
    # Skip non-string entries and macro entries. Macros themselves don't
    # need to be processed, and processing these can introduce problems
    # with recursion
    if not isinstance(tag.content, str):
        return None

    # Process the entry in the context
    tag.content = replace_macros(tag.content, tag.context)
    return tag



from .label import Label


class DocumentLabel(Label):
    """A label for documents.
    """
    title = None

    def __init__(self, doc_id, id, kind, title, order=None):
        super().__init__(doc_id=doc_id, id=id, kind=kind, order=order)
        self.title = title



"""
Attributes are modifiers to tags that change *how* a tag is rendered.
Attributes are represented by ordered dicts that can validate their entries.
"""
from .attributes import Attributes, AttributeFormatError

__all__ = ('Attributes', 'AttributeFormatError')



"""
A TargetBuilder for xhtml files.
"""
from .html_builder import HtmlBuilder


class XHtmlBuilder(HtmlBuilder):
    """A builder for xhtml files."""

    available = True
    priority = 1000
    infilepath_ext = '.dm'
    outfilepath_ext = '.xhtml'



"""
Path objects (pathlib.Path) that can keep track of subpaths for a project and
targets.
"""

from .paths import SourcePath, TargetPath
from . import utils

__all__ = ('SourcePath', 'TargetPath', 'utils')



"""Utilities for contexts."""
import regex

from ..utils.string import str_to_dict


re_header_block = regex.compile(r'^[\s\n]*(-{3,})\s*\n'
                                r'(?P<header>.+?)'
                                r'(\n\s*\g<1>)\n?', regex.DOTALL)


def find_header_entries(context):
    """Find context entries that contain a header string.

    Parameters
    ----------
    context : dict
        The context dict to search.

    Returns
    -------
    keys
        The keys for entries with header entries.
    """
    keys = set()
    for k, v in context.items():
        if isinstance(v, str) and re_header_block.match(v):
            keys.add(k)
    return keys


def load_from_string(string):
    """Load a context from a string containing a header string.

    Parameters
    ----------
    string : str
        The string to load into this context.

    Returns
    -------
    string, dict : Tuple[str, dict]
       The processed string with the header removed and the dict with the
       loaded values from the header.
    """
    # Pull out the header string, if available
    m = re_header_block.match(string)
    rest = None
    if m is not None:
        # Determine the start and end position of the header
        start, end = m.span()

        # Collect the rest of the string
        rest = string[end:]

        # Get the header part of the string
        string = m.groupdict()['header']

    # Parse the string
    d = str_to_dict(string)

    return rest, d



"""Bases classes for checkers of software dependencies."""


class SoftwareDependency(object):
    """An external software program or package dependency.

    Parameters
    ----------
    name : str
        The name of the external program or package.

    Attributes
    ----------
    path : str
        The path for the external program or package.
    available : Optional[bool]
        True, if dependency is available, False otherwise.
        May also be None if the availability is not determined yet
    """

    __slots__ = ('name', 'path', 'available')

    def __init__(self, name, **kwargs):
        """Initialize dependency with name and attributes."""
        kwargs['name'] = name
        if 'available' not in kwargs:
            kwargs['available'] = None

        for k, v in kwargs.items():
            if k in self.__slots__:
                setattr(self, k, v)

    def __repr__(self):
        return "{}({})".format(self.name,
                               self.available
                               if isinstance(self.available, bool) else
                               'unknown')


class SoftwareDependencyList(object):
    """A listing of software dependencies

    Parameters
    ----------
    category : str
        The category name for the software dependency list.
    *dependencies : Tuple[Union[str, :obj:`SoftwareDependencyList`]
        The listing of dependency names or sub dependency lists.
    """

    __slots__ = ('category', 'dependencies')

    def __init__(self, category, *dependencies):
        """Initialize with category and sub-dependencies."""
        self.category = category
        # if (len(dependencies) == 1 and
        #    isinstance(dependencies[0], SoftwareDependencyList)):
        #     self.dependencies = dependencies[0]
        # else:
        self.dependencies = [SoftwareDependency(i) if isinstance(i, str)
                             else i for i in dependencies]

    def __getitem__(self, key):
        if isinstance(key, str):
            for item in self.dependencies:
                if (getattr(item, 'name', None) == key or
                   getattr(item, 'category', None) == key):
                    return item
            raise KeyError(key)
        else:
            return self.dependencies[key]

    def __len__(self):
        return len(self.dependencies)

    def __contains__(self, key):
        if isinstance(key, str):
            for item in self.dependencies:
                if (getattr(item, 'name', None) == key or
                   getattr(item, 'category', None) == key):
                    return True
        return False

    def __repr__(self):
        cls_name = self.__class__.__name__
        items_str = ', '.join([i.__repr__() for i in self.dependencies])
        return cls_name + '[' + items_str + ']'

    def flatten(self, items=None, level=1):
        """Return a flattened list of software dependency objects.

        The flattened list includes SoftwareDependency and
        SoftwareDependencyList objects.

        Returns
        -------
        flattened_list : List[Tuple[int, Union[:obj:`.SoftwareDependency`, \
            :obj:`.SoftwareDependencyList`]]]
            A flattened list of tuples with the level (int) and software
            dependencies or software dependency lists.
        """
        # Recursion check
        if level > 16:
            return None
        flattened_list = []

        # Use this list's dependencies list, if not items are specified
        items = self if items is None else items
        items = items.dependencies if hasattr(items, 'dependencies') else items

        # Wrap in a list
        items = [items] if not isinstance(items, list) else items

        for item in items:
            # Only return SoftwareDependency or SoftwareDependencyList objects
            if isinstance(item, SoftwareDependency):
                flattened_list.append((level, item))
            elif isinstance(item, SoftwareDependencyList):
                flattened_list.append((level, item))
                flattened_list += self.flatten(items=item, level=level + 1)

        return flattened_list

    def keys(self):
        """Name or category of sub-dependencies."""
        return [i.category if hasattr(i, 'category')
                else i.name
                for i in self.dependencies]

    @property
    def available(self):
        return False


class Any(SoftwareDependencyList):
    """A list for external programs and packages in why any of them may
    be present."""

    @property
    def available(self):
        return any(i.available is True for i in self.dependencies)


class All(SoftwareDependencyList):
    """A list for external programs and packages in why all of them
    should be present."""

    @property
    def available(self):
        return all(i.available is True for i in self.dependencies)


class Optional(SoftwareDependencyList):
    """A list for external programs and packages in why none of them are
    required to be present."""

    @property
    def available(self):
        return True



"""
Exceptions for documents
"""


class DocumentException(Exception):
    """An error was encountered in creating, setting up or rendering a
    document."""
    pass


class TargetNotFound(DocumentException):
    """An error raised if a document target was requested but not found."""
    pass



"""
Exceptions for labels
"""


class LabelError(Exception):
    """An error was encountered while processing a label."""
    pass


class DuplicateLabel(LabelError):
    """A label that was already defined is defined again"""
    pass


class LabelNotFound(LabelError):
    """Could not find a reference to a label"""
    pass



"""
Static file handlers
"""
from tornado.web import StaticFileHandler

from .server import ServerHandler


class CustomStaticFileHandler(ServerHandler, StaticFileHandler):
    """A custom static file handler that handles rendered error pages"""
    pass



"""
Text formatting tags
"""
from html.entities import name2codepoint

from .tag import Tag
from .exceptions import TagError, assert_content_str
from ..formats import tex_cmd, tex_env, tex_verb, xhtml_entity, xhtml_tag
from .utils import format_content
from .. import settings


class Body(Tag):
    """A body tag for the body of a document.

    In (x)html, this is rendered as a <div> instead of the default <span>
    """

    active = True
    html_name = 'div'

    def html_fmt(self, attributes=None, **kwargs):
        attributes = attributes or self.attributes.copy()
        attributes['class'] = 'body'
        return super().html_fmt(attributes=attributes, **kwargs)

    def xhtml_fmt(self, content=None, attributes=None, format_func='xhtml_fmt',
                  method='xhtml', level=1, **kwargs):
        # Wrap the body tag in the epub namespace
        name = (self.html_name if self.html_name is not None else
                self.name.lower())

        # Collect the content elements
        content = content if content is not None else self.content
        content = format_content(content=content, format_func=format_func,
                                 level=level + 1)

        # Set the attributes
        attributes = attributes or self.attributes.copy()
        attributes['class'] = 'body'

        # Format the html tag
        return xhtml_tag(name=name, level=level, attributes=attributes,
                         formatted_content=content, method=method,
                         nsmap=settings.xhtml_namespace)


class P(Tag):
    """A Paragraph tag

    Attributes
    ----------
    active : bool
        This tag is active.
    include_paragraphs : bool
        The contents of this tag can be included in paragraphs.
    """
    active = True
    include_paragraphs = False

    def tex_fmt(self, **kwargs):
        tex = super(P, self).tex_fmt(**kwargs)

        # Rewrap the text
        # if settings.tex_paragraph_width > 0:
        #     tex = "\n".join(wrap(tex, settings.tex_paragraph_width))

        # Add newlines around paragraph if the content isn't a single tag
        # that is also a block
        if getattr(self.content, 'tex_paragraph_newlines', True):
            return "\n" + tex + "\n"
        else:
            return tex


class Bold(Tag):
    """A bold tag.

    Attributes
    ----------
    aliases : Tuple[str]
        A list of strs for other names a tag goes by
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    tex_cmd : str
        Use this name to render the tex command.
    active : bool
        This tag is active.
    """
    aliases = ("b", "textbf", "strong")
    html_name = "strong"
    tex_cmd = "textbf"
    active = True


class Italics(Tag):
    """An italics tag.

    Attributes
    ----------
    aliases : Tuple[str]
        A list of strs for other names a tag goes by
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    tex_cmd : str
        Use this name to render the tex command.
    active : bool
        This tag is active.
    """
    aliases = ("i", "textit")
    html_name = "i"
    tex_cmd = "textit"
    active = True


class Sup(Tag):
    """A superscript tag.

    Attributes
    ----------
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    active : bool
        This tag is active.
    """
    html_name = "sup"
    active = True

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        content = content if content is not None else self.content

        # Collect the content elements
        content = format_content(content=content, format_func='tex_fmt',
                                 level=level + 1, mathmode=mathmode)
        content = ''.join(content) if isinstance(content, list) else content

        content = '^{' + content + '}'
        return (tex_cmd(cmd='ensuremath', attributes='',
                        formatted_content=content)
                if not mathmode else
                content)


class Sub(Tag):
    """A subscript tag.

    Attributes
    ----------
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    active : bool
        This tag is active.
    """
    html_name = "sub"
    active = True

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        content = content if content is not None else self.content

        # Collect the content elements
        content = format_content(content=content, format_func='tex_fmt',
                                 level=level + 1, mathmode=mathmode)
        content = ''.join(content) if isinstance(content, list) else content

        content = '_{' + content + '}'
        return (tex_cmd(cmd='ensuremath', attributes='',
                        formatted_content=content)
                if not mathmode else
                content)


class Supsub(Tag):
    """A superscript/subscript tag, together, that displays them one on top of
    the other.

    The content of the tag consists of two elements separated by a '&&'
    character. ex: @supsub{superscript && subscript}

    Attributes
    ----------
    active : bool
        This tag is active.
    """
    active = True

    _sup = None
    _sub = None

    def __init__(self, *args, **kwargs):
        super(Supsub, self).__init__(*args, **kwargs)

        # Raise a TagError if this tag is nested."""
        if not isinstance(self.content, str):
            msg = "The @supsub tag cannot have tags nested within it."
            raise TagError(msg)

        # assert that the element separator, '&&', is in the contents
        if self.content.count('&&') != 1:
            msg = ("The @supsub tag must contain only one element '&&' "
                   "element separator. "
                   "ex: @supsub{{superscript && subscript}}. The "
                   "tag given was: {}")
            raise TagError(msg.format(self))

        # Seperate the superscript and subscript
        sup, sub = self.content.split('&&')
        self._sup = sup.strip()
        self._sub = sub.strip()

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1, **kwargs):
        attrs = attributes or self.attributes.copy()
        attrs['class'] = 'supsub'
        br_tag = xhtml_tag(name='br', method=method, level=level + 1)
        return xhtml_tag(name='span', attributes=attrs,
                         formatted_content=[self._sup, br_tag, self._sub],
                         method=method, level=level)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        formatted = "^{" + self._sup + "}_{" + self._sub + "}"
        return (tex_cmd(cmd='ensuremath', attributes='',
                        formatted_content=formatted)
                if not mathmode else
                formatted)


class Symbol(Tag):
    """One or more greek characters.

    Attributes
    ----------
    aliases : Tuple[str]
        A list of strs for other names a tag goes by
    active : bool, default: True
        This tag is active.
    """

    aliases = ("smb",)
    active = True

    def __init__(self, name, content, attributes, context):
        assert_content_str(content)
        super().__init__(name=name, content=content, attributes=attributes,
                         context=context)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        content = content if content is not None else self.content
        content = "\\" + content
        return (tex_cmd(cmd='ensuremath', attributes='',
                        formatted_content=content)
                if not mathmode else
                content)

    def html_fmt(self, content=None, attributes=None, method='html', level=1,
                 **kwargs):
        content = content or self.content
        return xhtml_entity(entity=content, method=method, level=level)

    def xhtml_fmt(self, content=None, attributes=None, method='xml', level=1,
                  **kwargs):
        content = (content or self.content).strip()

        # Convert the entity name to a code point, needed for epub.
        # ex: 'gamma' -> '#947'
        content = ('#' + str(name2codepoint[content])
                   if content in name2codepoint else content)
        return xhtml_entity(entity=content, method=method, level=level)


class Verb(Tag):
    """A verbatim tag for displaying unformatted blocks of text.

    Attributes
    ----------
    aliases : Tuple[str]
        A list of strs for other names a tag goes by
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    active : bool
        This tag is active.
    process_content : bool
        Do not process the contents of the tag--just take the contents
        literally.
    include_paragraphs : bool
        The contents of this tag cannot be included in paragraphs.
    """

    aliases = ("v", "pre", "verbatim")
    active = True

    process_content = False
    process_typography = False

    include_paragraphs = False

    html_name = 'code'

    def html_fmt(self, attributes=None, *args, **kwargs):
        attrs = attributes or self.attributes.copy()

        if self.name == "verbatim":
            attrs['class'] = 'block'
        return super(Verb, self).html_fmt(*args, **kwargs)

    def tex_fmt(self, *args, **kwargs):
        if self.name == "verbatim":
            return tex_env(env='verbatim', attributes='',
                           formatted_content=self.default_fmt())
        else:
            return tex_verb(self.default_fmt())



"""
Utilities for testing.
"""


def strip_leading_space(string, no_spaces=4):
    """Strip leading spaces at the start of newlines in a string.
    """
    pieces = [s[no_spaces:] if s.startswith(" " * no_spaces) else s
              for s in string.split('\n')]
    return '\n'.join(pieces)



"""
Objects to manage builds
"""
import logging
import pathlib
from abc import ABCMeta
from string import Formatter
from distutils.spawn import find_executable

import pathvalidate

from .executor import executor, run, runtime_error, runtime_success
from .utils import generate_outfilepath, generate_mock_parameters
from .exceptions import BuildError
from ..signals import signal
from ..utils.classes import all_subclasses
from ..utils.list import uniq, flatten
from ..paths import TargetPath
from .. import settings


class CustomFormatter(Formatter):
    """A custom formatter class for preparing actions into command-line
    arguments."""

    def get_field(self, field_name, args, kwargs):
        field_value, field_name = super().get_field(field_name, args, kwargs)
        if isinstance(field_value, list) or isinstance(field_value, tuple):
            # Make lists/tuples into space-separated strings
            field_value = " ".join(self.clean_field(f) for f in field_value)
            return field_value, field_name
        return self.clean_field(field_value), field_name

    @staticmethod
    def clean_field(f):
        """Clean fields used for command-line processes"""
        if isinstance(f, pathlib.Path):
            f = pathvalidate.sanitize_filepath(f, platform='auto')
        return str(f).strip('-*`')


class Builder(metaclass=ABCMeta):
    """A build for an output file.

    Parameters
    ----------
    env: :obj:`.builders.Environment`
        The build environment
    parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including filepaths, for the build
    outfilepath : Optional[:obj:`pathlib.Path`]
        If specified, the path for the output file. If not specified, an
        outfilepath will be automatically generated.
    use_cache : Optional[bool]
        If True, set the builder outfilepath into the cache_path from the
        builder environment. Note that this will also place temporary files
        created by the builder in the same directory.
    use_media : Optional[bool]
        If True, set the builder outfilepath subpath in the media_path in the
        build environment context, if specified.

    Attributes
    ----------
    action : str
        The command to execute during the build.
    available : bool
        Whether this builder is available to factory methods (find_builder_cls)
    active_requirements : Union[tuple, bool]
        If False, the builder will be inactive
        If a tuple of strings is specified, these conditions will be tested
        to see if the builder is active:

          - 'priority': test that the priority attribute is an int
          - 'required_execs': tests that the required_execs attribute is
             specified
          - 'all_execs': tests that the required execs are available
    decision : :obj:`.builders.deciders.decider.decision`
        The decision object for the build, instantiate from environment's
        decider, to evaluate whether a build is needed.
    scan_parameters_on_init : bool
        If True (default), scan the parameters for additional dependencies
        during the __init__.
    priority : int
        If multiple viable builders are available, use the one with the highest
        priority.
    required_execs : Tuple[str]
        A list of external executables that are needed by the builder.
    infilepath_ext : str
        The format extension for the input parameters (ex: '.pdf', '.render')
    outfilepath_ext : str
        The format extension for the output file (ex: '.svg')
    outfilepath_append : str
        For automatically generated outfilepaths, the following string will
        be appended to the name of the file. ex: '_scale'
    target : Optional[str]
        If specified, use the given document target for the build. This is used
        in formatting the TargetPath.
        ex: 'html' target will store built files in the 'html/' subdirectory.
    parameters_from_signals : Optional[List[str]]
        A list of optional signal names to receive extra parameter
        dependencies.
    future : Union[:obj:`concurrent.futures.Future`, None, str]
        The future object for the process for the externally run program.
        The future can also be None, if a process hasn't been run.
    """
    env = None
    action = None
    available = False
    active_requirements = ('priority', 'required_execs', 'all_execs')
    decision = None
    scan_parameters_on_init = True

    priority = None
    required_execs = None
    use_cache = False
    use_media = True

    infilepath_ext = None
    outfilepath_ext = None

    # Options that impact how the outfilepath is formatted
    outfilepath_append = None
    target = None

    parameters_from_signals = None

    future = None
    timeout = settings.default_timeout

    _active = dict()
    _available_builders = dict()
    _parameters = None
    _missing_parameters = None
    _outfilepath = None

    runtime_error = runtime_error  # handler function when runtime error found
    runtime_success = runtime_success  # handler to test succesful run

    def __init__(self, env, target=None, parameters=None, outfilepath=None,
                 use_cache=None, use_media=None, **kwargs):
        self.env = env
        self.target = target.strip('.') if isinstance(target, str) else None
        self.use_cache = use_cache if use_cache is not None else self.use_cache
        self.use_media = use_media if use_media is not None else self.use_media

        # Load the parameters. The parameters should be a list, and if a
        # parameters list was passed, make a copy of this list.
        parameters = parameters if parameters is not None else []
        parameters = (list(parameters) if isinstance(parameters, tuple) or
                      isinstance(parameters, list) else [parameters])
        self.parameters = parameters

        # Scan for additional parameters, if desired
        if self.scan_parameters_on_init:
            self.scan_parameters()

        # Load the outfilepath
        self.outfilepath = (outfilepath if isinstance(outfilepath, TargetPath)
                            else None)

    def __repr__(self):
        return "<{} status='{}'>".format(self.__class__.__name__,
                                         self.status)

    @classmethod
    def active(cls):
        """True if a builder is active"""
        cls_name = cls.__name__

        # See if the specified class has already been evaluated to be active or
        # not
        if cls_name in Builder._active:
            return Builder._active[cls_name]

        active = True

        # 1. Make sure the priority is properly set for the concrete class
        if (cls.active_requirements and
            'priority' in cls.active_requirements and
           not isinstance(cls.priority, int)):
            logging.warning("Builder '{}' not active because a priority has "
                            "not been set.".format(cls.__name__))
            active = False

        # 2. Make sure the required executable tuple is properly set for the
        #    concrete class
        required_execs = isinstance(cls.required_execs, tuple)
        if (cls.active_requirements and
            'required_execs' in cls.active_requirements and
           not required_execs):
            logging.warning("Builder '{}' not active because the required "
                            "executable tuple was not "
                            "specified.".format(cls.__name__))
            active = False

        # 3. Make sure the required executables can be found
        if (cls.active_requirements and
            'all_execs' in cls.active_requirements and
           cls.required_execs):
            all_execs = {exe: find_executable(exe)
                         for exe in cls.required_execs}
            if not all(v is not None for v in all_execs.values()):
                missing_execs = [exe for exe, available in all_execs.items()
                                 if available is None]
                logging.warning("Builder '{}' not active because the "
                                "required executables could not be "
                                "found: {}".format(cls.__name__,
                                                   missing_execs))
                active = False

        return Builder._active.setdefault(cls_name, active)

    @property
    def status(self):
        """The status of the builder.

        The builder can have the following states:
          - 'ready': The builder is active and the parameters have been set
          - 'inactive': The builder isn't active--see the active property
          - 'missing (parameters)': All the required parameters have not been
            specified or files for paths in the parameters do not exist
          - 'missing (outfilepath)': The outfilepath was not created
          - 'cancelled' : The build was cancelled.
          - 'building': The builder is building
          - 'done': The builder is done building
        """
        active = self.active

        if not active:
            return "inactive"

        if self.missing_parameters:
            return "missing (parameters)"

        # If a build is not needed or the process is done, then the build is
        # done
        if not self.build_needed():
            return "done"

        # If a process is open (self.open is not None), then a build is
        # currently in process
        if self.future is not None:
            if not self.future.done():
                # The process isn't done.
                return "building"

            if self.future.cancelled():
                return "cancelled"

            exc = self.future.exception()
            if exc is not None:
                # Raise the exception, if an exception was raised
                raise exc

            run_successful = self.runtime_success(future=self.future)
            if run_successful:
                # Couldn't find the generated file!
                if not self.outfilepath.exists():
                    cls_name = self.__class__.__name__
                    msg = ("The '{}' build was successful but could not find "
                           "the output file '{}'")
                    logging.error(msg.format(cls_name, self.outfilepath))

                    return "missing (outfilepath)"

                # It was a successful run and the generated file exists
                else:
                    self.build_needed(reset=True)
                    return "done"

            # non-zero exit code. Unsuccessful. :(
            self.runtime_error(future=self.future)

        else:
            return "ready"

    def build_needed(self, reset=False):
        """Decide whether a build is needed"""
        if self.decision is None:
            decider = self.env.decider
            self.decision = decider.decision
        inputs = list(self.parameters)
        if self.action:
            inputs.append(self.action)
        return self.decision.build_needed(inputs=inputs,
                                          output=self.outfilepath,
                                          reset=reset)

    @property
    def parameters(self):
        """The list of input parameters, including filepaths, needed for the
        build"""
        parameters = list(getattr(self, '_parameters', []))
        parameters += self.get_parameters_from_signals()

        # Input parameters should only be listed once.
        return uniq(parameters)

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    @property
    def missing_parameters(self):
        """Returns True if there are no parameters or there are missing
        parameters.

        This function will check the parameters until the parameters are all
        found, in which case, it will just return False.
        """
        if self._missing_parameters is False:
            return False

        parameters = self.parameters
        has_parameters = len(parameters) > 0

        if not has_parameters:
            return True
        elif not all(i.exists() for i in parameters if hasattr(i, 'exists')):
            return True
        else:
            self._missing_parameters = False
            return False

    def get_parameter(self, name, *parameters):
        """Return parameters inserted as 2-ples of name/value.

        Parameters
        ----------
        name
            The name of the 2-ple parameter (the 1st 2-ple item)
        *parameters
            If specified, use these parameters to get the specified parameter.
            Otherwise, the builder's parameters will be used.

        Returns
        -------
        value
            The corresponding 2-ple value (the 2nd 2-ple item)
        """
        parameters = parameters or self.parameters or []
        filtered_parameters = [p[1] for p in parameters
                               if isinstance(p, tuple) and len(p) > 1 and
                               p[0] == name]

        return filtered_parameters[0] if filtered_parameters else None

    def get_parameters_from_signals(self, sort=True):
        """Retrieve additional parameters from emitted signals specified in
        the parameters_from_signals list attribute."""
        parameters = []

        if isinstance(self.parameters_from_signals, str):
            self.parameters_from_signals = [self.parameters_from_signals]

        if isinstance(self.parameters_from_signals, list):
            # Retrieve the signals
            for signal_name in self.parameters_from_signals:
                # Convert the signal name to an actual signal
                sig = signal(signal_name)
                rv = sig.emit(builder=self)
                parameters += list(flatten(rv))

        return list(sorted(parameters)) if sort else parameters

    def scan_parameters(self):
        """Use the environment scanners to find additional dependencies in
        files specified by filepaths in the parameters."""
        parameters = self.parameters
        parameters += self.env.scanner.scan(parameters=parameters)
        self.parameters = uniq(parameters)

    @property
    def infilepaths(self):
        """Retrieve the parameters that are filepaths (:obj:`pathlib.Path`)."""
        return [p for p in self.parameters if isinstance(p, pathlib.Path)]

    @property
    def not_infilepaths(self):
        """Retrieve the parameters that are not filepaths (:obj:`pathlib.Path`)
        """
        return [p for p in self.parameters if not isinstance(p, pathlib.Path)]

    @property
    def outfilepath(self):
        """The output filename and path"""
        outfilepath = self._outfilepath

        if outfilepath is None:
            # Generate the parameters for the generate_outfilepath
            infilepaths = self.infilepaths
            if infilepaths:
                # Use the filepaths for the input files, if available
                parms = infilepaths
            else:
                # Otherwise generate mock filepath parameters
                context = getattr(self, 'context', None)
                parms = generate_mock_parameters(env=self.env,
                                                 context=context,
                                                 parameters=self.parameters,
                                                 ext=self.infilepath_ext)

            # Use the given parameters to generate a filepath for the output
            # file
            outfilepath = generate_outfilepath(env=self.env,
                                               parameters=parms,
                                               target=self.target,
                                               append=self.outfilepath_append,
                                               ext=self.outfilepath_ext,
                                               use_cache=self.use_cache,
                                               use_media=self.use_media)

        # Make sure the outfilepath directory exists
        if outfilepath and not outfilepath.parent.is_dir():
            outfilepath.parent.mkdir(parents=True, exist_ok=True)

        self._outfilepath = outfilepath
        return outfilepath

    @outfilepath.setter
    def outfilepath(self, value):
        self._outfilepath = value

    def run_cmd_args(self):
        """Format the action, if it's a string.

        Returns
        -------
        run_cmd_args : Tuple[str]
            A tuple of the arguments to run in a process.
        """
        if isinstance(self.action, str):
            fmt = CustomFormatter()
            fmt_action = fmt.format(self.action, builder=self)
            return tuple(fmt_action.split())
        else:
            return tuple()

    def run_cmd(self, *args):
        """If the action is a external command, run it."""
        if self.future is None and (isinstance(self.action, str) or args):

            # Format the action string, if it's to be used
            args = args if args else self.run_cmd_args()
            logging.debug("'{}' run with: '{}'".format(self.__class__.__name__,
                                                       " ".join(args)))

            # add the process to the executor pool
            future = executor.submit(run, args=args, timeout=self.timeout)
            self.future = future

    def build(self, complete=False):
        """Run the build.

        Parameters
        ----------
        complete : Optional[bool]
            If True, run the build until it has completed
            If False, start the build in the background.

        Returns
        -------
        status : str
            The current status of the build.

        .. note::
          - This function will run the sub-builders.
          - Each builder is atomic
          - When running a build, not all of the builders might be called in
            the first build--for example, subsequent builders may rely on the
            results of previous builds. For this reason, builders should be
            used in conjunction with an environment to make sure a set of
            builds are completed or the build complete=True should be used.
        """
        if complete:
            # Run while this builder is either ready to build or a build is
            # ongoing.
            while self.status in {'building', 'ready'}:
                self.run_cmd()
        else:
            if self.status in {'building', 'ready'}:
                self.run_cmd()
        return self.status

    @classmethod
    def find_builder_cls(cls, in_ext, out_ext=None, target=None,
                         raise_error=True):
        """Factory method to return a builder class

        Parameters
        ----------
        in_ext: str
            The extension for the input file.
        out_ext : str
            The extension for the output file.
        target : Optional[str]
            The document target. ex: '.html' or '.pdf'
        raise_error : Optional[bool]
            If True (default), raise an exception if a builder class couldn't
            be found.
        """
        # Cache the available builders
        if not Builder._available_builders:
            subclses = Builder._available_builders

            for builder in all_subclasses(Builder):
                logging.debug("Builder {:<30}: active={}, available={}"
                              "".format(builder.__name__, builder.active(),
                                        builder.available))

                if not builder.available or not builder.active():
                    continue

                key = (builder.infilepath_ext, builder.outfilepath_ext)

                # Only replace an existing builder if it's higher property
                if (key not in subclses or
                   subclses[key].priority < builder.priority):
                    subclses[key] = builder

        # See if there's a valid builder from the given out_ext
        key = (in_ext, out_ext)
        if target is None and key in Builder._available_builders:
            return Builder._available_builders[key]

        # Otherwise see if there's a tracked target
        target = (target if not isinstance(target, str) or
                  target.startswith('.') else '.' + target)

        if target in settings.tracked_deps:
            # If the in_ext is an allowed format, just use a copy builder
            # defined as the builder with a '.*' infilepath_ext and
            # outfilepath_ext
            if in_ext in settings.tracked_deps[target]:
                copy_builder = Builder._available_builders[('.*', '.*')]
                return copy_builder

            # Otherwise find a converter
            for out_ext in settings.tracked_deps[target]:
                key = (in_ext, out_ext)
                if key in Builder._available_builders:
                    return Builder._available_builders[key]

        # No builder class could be found
        if raise_error:
            msg = ("A builder cannot be found for '{}'. Available formats "
                   "are: {}".format(in_ext, settings.tracked_deps))
            raise BuildError(msg)
        else:
            return None



"""
The pool executor for running multiple functions at once.
"""
import logging
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from collections import namedtuple
import subprocess

from .exceptions import BuildError

# Setup a global pool for processes
executor = ThreadPoolExecutor(max_workers=cpu_count() * 2)

PopenResult = namedtuple('PopenResult', 'returncode args stdout stderr')


def run(timeout, **kwargs):
    """Run the command with the given arguments."""
    popen = subprocess.Popen(**kwargs, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, bufsize=4096,)
    popen.wait(timeout=timeout)
    stdout, stderr = popen.communicate()
    return PopenResult(returncode=popen.returncode,
                       args=popen.args,
                       stdout=stdout.decode('latin1'),
                       stderr=stderr.decode('latin1'))


@staticmethod
def runtime_success(future):
    """Test whether a future from a subprocess is successful."""
    return future.result().returncode == 0


@staticmethod
def runtime_error(future, error_msg=None, raise_error=True):
    """Raise an error from a future working with subprocess"""
    popen_result = future.result()
    returncode = popen_result.returncode
    args = popen_result.args
    out = popen_result.stdout
    err = popen_result.stderr

    if error_msg is None:
        error_msg = ("The conversion command '{}' was "
                     "unsuccessful. Exited with code "
                     "{}.".format(' '.join(args), returncode))
    if raise_error:
        e = BuildError(error_msg)
        e.cmd = " ".join(args)
        e.returncode = returncode
        e.shell_out = out
        e.shell_err = err
        raise e
    else:
        logging.warning(error_msg)
        logging.debug(err)



"""
The Label tag to reference captions and other labels.
"""
from .tag import Tag
from .utils import content_to_str
from .exceptions import assert_content_str
from .utils import format_content
from ..attributes import Attributes
from ..formats import xhtml_tag
from ..utils.string import titlelize, slugify
from ..utils.classes import weakattr


def generate_label_id(tag):
    """Generate a label_id string for the given tag.

    Parameters
    ----------
    tag : :obj:`Tag <.Tag>`
        The tag creating the label.

    Returns
    -------
    label_id : str
        The label id string for the tag.
    """
    assert tag.context.is_valid('doc_id')
    context = tag.context
    doc_id = context['doc_id']
    # Retrieve the tag attributes and content, and make sure they're in the
    # needed formats
    content = tag.content

    attributes = tag.attributes
    attributes = (Attributes(attributes) if isinstance(attributes, str)
                  else attributes)

    if 'id' in attributes:
        # If an 'id' was specified, just use that.
        return attributes['id']
    elif 'short' in attributes:
        # Convert the 'short' title to a title for the label and slugify to
        # remove spaces
        title = titlelize(attributes['short'])
        title = '-'.join((doc_id, title))
        return slugify(title)

    # Try to generate the label_id from the contents
    elif isinstance(content, str) and content.strip() == '':
        # The contents of the tag are empty. Make an automatically generated
        # label
        label_count = context.setdefault('label_count', 0) + 1
        context['label_count'] = label_count
        doc_id = slugify(context['doc_id'])

        return "-".join((doc_id, str(label_count)))
    else:
        # Convert the content to a title for the label and slugify to
        # remove spaces
        title = titlelize(content_to_str(content))
        title = '-'.join((doc_id, title))
        return slugify(title)


def create_label(tag, kind, label_id=None):
    """Create a label in the label_manager for the given tag.

    .. note:: This function does not create a :obj:`LabelTag
              <disseminate.tags.label.LabelTag>`. Only a :obj:`Label
              <disseminate.labels.labels.Label>` object through the document's
              :obj:<LabelManager
              <disseminate.labels.label_manager.LabelManager>`

    Parameters
    ----------
    tag : :obj:`Tag <disseminate.tags.tag.Tag>`
        The tag creating the label.
    kind : Tuple[str]
        The kind of label. ex: ('heading', 'section')
    label_id : Optional[str]
        The identifier for the label to create.
        If None is specified, the generate_label_id function will be used.

    Returns
    -------
    label_id : Union[str, None]
        The label_id of the created label, if successful.
        None if a label could not be created.
    """
    content = tag.content
    context = tag.context

    assert 'label_manager' in context, ("A label manager could not be found "
                                        "in the tag's context")

    # Generate the label_id, if needed
    label_id = label_id if label_id is not None else generate_label_id(tag=tag)

    # Get the tag's context and attributes
    attributes = tag.attributes
    attributes = (Attributes(attributes) if isinstance(attributes, str) else
                  attributes)

    # Form a title from the content
    if 'short' in attributes:
        content = attributes['short']
    else:
        content = titlelize(content_to_str(content))

    # Create the content label
    label_manager = context['label_manager']
    label_manager.add_content_label(id=label_id, kind=kind,
                                    title=content, context=context)
    return label_id


class LabelMixin(object):
    """A mixin class for working with LabelTags LabelAnchors.

    This mixing works in coordination with the :class:`Tag
    <disseminate.tags.tag.Tag>` base class. The constructor (__init__) of this
    class should be run after the Tag base class's constructor.
    """

    label_id = None
    label_tag = None
    label_anchor = None

    def __init__(self, *args, **kwargs):
        self.create_label()

    def generate_label_id(self):
        """Generate the label_id to use in creating the label.

        Override this function to customize the generation of the label_id.

        Returns
        -------
        label_id : str
            The generated label_id.
        """
        if 'id' in self.attributes:
            # Get the label_id from the attributes, if specified.
            return self.attributes['id']
        else:
            # No label id specified. Generate a label_id
            return generate_label_id(tag=self)

    def generate_label_kind(self):
        """Generate the kind tuple for the created label.

        Override this function to customize the generation of the label kind.

        Returns
        -------
        kind : Tuple[str]
            A tuple of strings for the kind of label.
            ex: ('heading', 'chapter')
        """
        return tuple()

    def create_label(self):
        """Create the label in the label_manager for this tag."""
        # Create the label. First,
        label_id = self.generate_label_id()
        kind = self.generate_label_kind()
        label_id = create_label(tag=self, kind=kind, label_id=label_id)

        # Add the label identifier to this tag's attributes. This will be
        # this tag's anchor for targets like html
        self.label_id = label_id
        self.attributes['id'] = label_id

        attrs = self.attributes.filter('id', 'class', 'short')
        context = self.context

        self.label_tag = LabelTag(name='label', attributes=attrs,
                                  content=label_id, context=context)
        self.label_anchor = LabelAnchor(name='label_anchor', attributes=attrs,
                                        content=label_id, context=context)


class LabelAnchor(Tag):
    """The label anchor."""

    name = 'label'
    tex_cmd = 'label'
    html_name = 'span'

    def __init__(self, *args, content, **kwargs):
        # Set the label_id from the content
        assert_content_str(content)
        label_id = content.strip()
        self.label_id = label_id

        super().__init__(*args, content=content, **kwargs)

    def html_fmt(self, method='html', level=1, **kwargs):
        # <span id="#label-id" />
        html_id = self.label_id
        return xhtml_tag('span', attributes='id=' + html_id, method=method,
                         level=level)


class LabelTag(Tag):
    """The name portion of a label.

    ex: Chap. 1.2 or Fig. 1.

    Parameters
    ----------
    content : str
        The label_id of an existing label.
    kind : Tuple[str]
        The kind of the label is a tuple that identified the kind of a label
        from least specific to most specific. ex: ('figure',), ('chapter',),
        ('equation',), ('heading', 'h1',)
        See :meth:`LabelManager.format_string
        <disseminate.labels.label_manager.LabelManager.format_string>`
    prepend_id : Optional[str]
        If specified, the label id (identifier) will have this string prepended
        to it.

    (see the :class:`Tag <disseminate.tags.tag.Tag>` base class for details
     on the other parameters)
    """

    # This tag should not be created by the tag factory; it should be created
    # by other tags.
    active = False

    label_id = None
    label_manager = weakattr()

    #: Do not convert typography characters. Conversion of typography
    #: characters might cause issues with matching the label id.
    process_content = False
    process_typography = False

    def __init__(self, name, attributes, content, context, **kwargs):
        # Set the label_id from the content
        assert_content_str(content)
        label_id = content.strip()
        self.label_id = label_id

        # Set the label_manager and get the format key
        if 'label_manager' in context:
            self.label_manager = context['label_manager']

        super().__init__(name=name, attributes=attributes,
                         content=content, context=context, **kwargs)

        # Clean up the attributes.
        # 1. Remove the 'id' attribute, since it will be included by the
        #    LabelAnchor
        if 'id' in self.attributes:
            del self.attributes['id']

    def default_fmt(self, content=None, attributes=None):
        # Get the label tag format
        label_manager = self.label_manager
        label_id = self.label_id
        context = self.context

        if all(i is not None for i in (label_manager, label_id, context)):
            format_str = label_manager.format_string(id=self.label_id)
            processed_tag = Tag(name='label', content=format_str,
                                attributes='', context=context)
            return content_to_str(processed_tag.content)
        else:
            return ''

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1):
        # Get the label tag format
        label_manager = self.label_manager
        label_id = self.label_id
        context = self.context

        if all(i is not None for i in (label_manager, label_id, context)):
            format_str = label_manager.format_string(id=self.label_id,
                                                     target='.tex')

            # Process the tags and format the contents for tex
            processed_tag = Tag(name='label', content=format_str,
                                attributes='', context=context)
            content = format_content(content=processed_tag.content,
                                     format_func='tex_fmt', level=level + 1,
                                     mathmode=mathmode)
            return ''.join(content) if isinstance(content, list) else content
        else:
            return ''

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1, **kwargs):
        # Get the label tag format
        label_manager = self.label_manager
        label_id = self.label_id
        context = self.context

        if all(i is not None for i in (label_manager, label_id, context)):
            # Retrieve the format string for the label
            # Use the 'html' format_str for html and xhtml
            format_str = label_manager.format_string(id=self.label_id,
                                                     target='.html')

            # Process the tags and format the contents for html (html_fmt) or
            # xhtml (xhtml_fmt)
            processed_tag = Tag(name='label', content=format_str,
                                attributes='', context=context)
            content = format_content(content=processed_tag.content,
                                     format_func=format_func, level=level + 1)

            attributes = (self.attributes.copy()
                          if attributes is None else attributes)
            attributes['class'] = 'label'
            return xhtml_tag('span', attributes=attributes,
                             formatted_content=content, method=method,
                             level=level)
        else:
            return ''

    def xhtml_fmt(self, format_func='xhtml_fmt', method='xhtml', **kwargs):
        return self.html_fmt(format_func=format_func, method=method, **kwargs)



"""
Composite builders for multiple build commands
"""

from .composite_builder import CompositeBuilder
from .sequential_builder import SequentialBuilder
from .parallel_builder import ParallelBuilder

__all__ = ('CompositeBuilder', 'SequentialBuilder', 'ParallelBuilder')



"""
Receivers for TargetBuilders
"""
from ..builder import Builder
from ...signals import signal

add_file = signal('add_file')
find_builder = signal('find_builder')
document_tree_updated = signal('document_tree_updated')
document_build = signal('document_build')
document_build_needed = signal('document_build_needed')


@add_file.connect_via(order=1000)
def add_file(parameters, context, in_ext, target, use_cache=False):
    """Add a file to the target builder.

    Parameters
    ----------
    parameters : Union[str, List[str, :obj:`pathlib.Path`]]
        The parameters for the build
    context : :obj:`.BaseContext`
        The document context dict
    in_ext : str
        The extension for the file.
    target : str
        The document target to add the file for.
    use_cache : Optional[bool]
        If True, use cached paths when adding files.

    Returns
    -------
    outfilepath : :obj:`pathlib.Path`
        The outfilepath for the file from the build.
    """
    # If not get the outfilepath for the given document target
    assert context.is_valid('builders')

    # Retrieve the unspecified arguments
    target = target if target.startswith('.') else '.' + target

    # Retrieve builder
    target_builder = context.get('builders', dict()).get(target)
    assert target_builder, ("A target builder for '{}' is needed in the "
                            "document context".format(target))

    build = target_builder.add_build(parameters=parameters,
                                     context=context,
                                     in_ext=in_ext,
                                     target=target,
                                     use_cache=use_cache)
    return build.outfilepath


# Loaded after headers but before tags are converted
@document_tree_updated.connect_via(order=2000)
def add_target_builders(root_document):
    """Add target builders to a document context"""
    # Get the subdocuments in the tree
    docs = root_document.documents_list(only_subdocuments=False,
                                        recursive=True)

    # Reset all of the builders
    for doc in docs:
        doc.context.setdefault('builders', dict()).clear()

    # Populate the target builders for each document
    for doc in docs:
        context = doc.context
        builders = context.setdefault('builders', dict())

        environment = context['environment']
        for target in context.targets:
            target = target if target.startswith('.') else '.' + target

            if target not in builders:
                builder_cls = Builder.find_builder_cls(in_ext='.dm',
                                                       out_ext=target)

                # Only add the target builder if not only_root or, if
                # only_root, the doc is equal to the root_document.
                if not builder_cls.only_root or doc == root_document:
                    builders[target] = builder_cls(env=environment,
                                                   context=context)


@find_builder.connect_via(order=1000)
def find_builder(context, target=None, **kwargs):
    """Find a target builder in a document context, or None if None was found.

    Parameters
    ----------
    context : :obj:`document.DocumentContext`
        The context dict for the document.
    target : Optional[str]
        The target to retrieve the target builder for.

    Returns
    -------
    builders : List[:obj:`.builders.target_builders.TargetBuilder`]
        Returns a list of all matching target builders.
    """
    assert 'builders' in context

    builders = context['builders']

    if target is not None:
        target = target if target.startswith('.') else '.' + target

        if target in builders:
            return [builders[target]]
        else:
            return []
    else:
        return list(builders.values())


# Loaded on document build
@document_build.connect_via(order=1000)
def build(document, complete=True):
    """Build a document tree's targets (and subdocuments) using the target
    builders."""

    # Setup a parallel builder
    context = document.context
    env = context['environment']
    root_builder = env.create_root_builder(document)

    if complete:
        while root_builder.status in {'building', 'ready'}:
            root_builder.build(complete=False)
    else:
        if root_builder.status in {'building', 'ready'}:
            root_builder.build(complete=False)

    return root_builder.status


@document_build_needed.connect_via(order=1000)
def build_needed(document):
    """Evaluate whether any of the target builders need to be build"""

    # Get the target builders
    # The builders dict should be in context from add_target_builders
    context = document.context

    assert 'builders' in context
    builders = context['builders']

    # check if any of the target builders need building
    return any(b.build_needed() for b in builders.values())



import weakref

from .exceptions import DuplicateSignal


class Signal(object):
    """A notification emitter.

    (Inspired by blinker)

    Attributes
    ----------
    receivers : Dict[int, Callable]
        A dict with the order (key) for a receiver (value) to run when the
        signal is emitted.
    """

    name = None
    receivers = None

    def __init__(self, name, doc=None):
        self.name = name
        if doc is not None:
            self.__doc__ = doc

        self.receivers = dict()

    def connect(self, receiver, order, weak=True):
        """Connect a receiver to this signal.

        Parameters
        ----------
        receiver : Callable
            A function that will be executed when this signal is emitted.
        order : int
            The order for this receiver to be run.
        weak : Optional[bool]
            If true, a weak reference will be stored for the receiver.
        """
        if order in self.receivers:
            msg = ("A receiver ({}) with order '{}' already exists in this "
                   "signal")

            raise DuplicateSignal(msg.format(self.receivers[order], order))
        self.receivers[order] = (weakref.ref(receiver) if weak else receiver)

    def connect_via(self, order, weak=True):
        """The decorator for connect"""
        def decorator(fn):
            self.connect(fn, order=order, weak=weak)
            return fn
        return decorator

    def emit(self, **kwargs):
        """Emit (send) the signal and run the receiver functions."""
        return_values = []
        for order, receiver in sorted(self.receivers.items()):
            # De-reference receiver, if needed
            receiver = (receiver() if isinstance(receiver, weakref.ref) else
                        receiver)
            if receiver is not None:
                return_values.append(receiver(**kwargs))
            else:
                return_values.append(None)
        return return_values

    def receivers_dict(self):
        """Return a dict of receivers (values) and their orders (keys)."""
        d = {order: rec() if isinstance(rec, weakref.ref) else rec
             for order, rec in self.receivers.items()}
        return {k: v for k, v in d.items() if v is not None}

    def reset(self):
        """Reset the signal to its initial state"""
        self.receivers.clear()


class Namespace(dict):
    """A mapping of signal names to signals."""

    def signal(self, name, doc=None):
        """Return a signal with the given name.

        Parameters
        ----------
        name : str
            The name of the signal
        doc : Optional[str]
            The description documentation for the signal.
        """
        try:
            signal = self[name]
            if doc is not None:
                signal.__doc__ = doc
            return signal
        except KeyError:
            return self.setdefault(name, Signal(name, doc))


signals = Namespace()
signal = signals.signal



"""
The 'init' CLI sub-command.
"""
import pathlib
from textwrap import TextWrapper, dedent
from shutil import copy2

import click

from .term import term_width, wrap_with_newlines
from ..utils.string import str_to_dict
from .. import settings


# Get the path for the templates directory
template_path = pathlib.Path(__file__).parent.parent / 'templates'


@click.command()
@click.argument('names', nargs=-1)
@click.option('--out-dir', '-o', required=False,
              type=click.Path(exists=True, file_okay=False, dir_okay=True),
              default=None,
              help="the directory to create the project starter")
@click.option('--info', is_flag=True, default=False,
              help="Show detailed information on the specified project "
                   "starter")
@click.option('-l', '--list', 'show_list',
              is_flag=True, default=False,
              help="List the available project starters")
def init(names=None, out_dir=None, info=False, show_list=False):
    """Initialize a new template project"""
    starters_dict = generate_starters_dict()

    if len(names) == 0 or show_list:
        print_starters_list(starters_dict)
        exit()

    name = names[0]

    # Make sure the starter name is listed in the starter dict
    if name not in starters_dict:
        msg = "The project starter with name '{}' could not be found."
        raise click.BadParameter(msg.format(name))

    # Retrieve the starter
    starter = starters_dict[name]

    # Print the detailed info, if specified
    if info:
        print_starter_info(name=name, starter=starter)
        exit()

    # At this stage, try cloning the project starter
    clone_starter(name=name, starter=starter, out_dir=out_dir)


def filepath_key(path):
    """A sort key function for paths"""
    parts = pathlib.Path(path).parts
    num_parts = len(parts)
    return num_parts, parts


def starter_path_to_name(relpath):
    """Convert a starter relative path (relative to template dir) to a
    starter name"""
    parts = relpath.parts

    # Remove the 'starter' directory name and the 'description.yaml' filename
    parts = [part for part in parts
             if part != settings.template_starter_dir and
             part != settings.template_starter_desc_filename]
    return '/'.join(parts)


def starter_path_template_name(relpath):
    """Extract the template name for a given starter path"""
    parts = relpath.parts

    # Find the 'starters' directory position in the parts tuple
    pos = parts.index(settings.template_starter_dir)

    return '/'.join(parts[0:pos])


def generate_starters_dict():
    """Create a dict of available starters with the starter name as the key
    and the full path for the starter as the value.

    Returns
    -------
    starters_dict : Dict[str, Dict[str,str]]
        A dict for all available project starters. The key is the starter name
        (str) and the value is a starter dict with the description of the
        starter.
    """
    # Find the configured starter directories in the template dir
    desc_filename = settings.template_starter_desc_filename
    description_filepaths = template_path.glob('**/' + desc_filename)

    starter_dict = dict()

    for path in description_filepaths:
        # Get the path relative to the template directory
        relpath = path.relative_to(template_path)

        # Generate the starter name
        name = starter_path_to_name(relpath)

        # Load the starter information
        starter = str_to_dict(path.read_text())

        # Workup the values
        starter['template'] = starter_path_template_name(relpath)
        starter['description'] = starter['description'].strip()
        starter['_relpath'] = relpath
        starter['_path'] = path

        # Generate a starter entry
        starter_dict[name] = starter

    return starter_dict


def starter_filepaths(starter):
    """Give a starter dict, return a dict with the filepaths for the files
    in the project starter.

    Parameters
    ----------
    starter : dict
        A dict with entries on the project starter. These are the entries
        (values) from the generate_starters_dict function.

    Returns
    -------
    filepaths : Dict[str, :obj:`pathlib.Path`]
        A dict with the filepaths for the project starter. The keys are
        the relative filepath strings, and the values are the full filepaths.
    """
    assert '_path' in starter
    starter_root = starter['_path'].parent

    filepaths = dict()
    for filepath in starter_root.glob('**/*'):
        if (filepath.name == settings.template_starter_desc_filename or
           not filepath.is_file()):
            continue

        relpath = filepath.relative_to(starter_root)
        filepaths[str(relpath)] = filepath

    return filepaths


def print_starters_list(starters_dict):
    """Print a list of available starters

    Parameters
    ----------
    starters_dict : Dict[str, Dict[str,str]]
        A dict for all available project starters. The key is the starter name
        (str) and the value is a starter dict with the description of the
        starter.
    """
    twidth = min(80, term_width())
    name_clr = settings.cli_init_starter_name_color

    # Normalize the string length of starter names
    max_len = max(len(name) for name in starters_dict.keys())

    # Setup the textwrapper
    wrap = TextWrapper(subsequent_indent=' ' * (max_len + 3),
                       width=twidth)

    # Print the starter names and short descriptions
    for name in sorted(starters_dict):
        starter = starters_dict[name]

        # textwrap the short description
        name_len = len(name)
        start = click.style(name, fg=name_clr)
        start += " " * (max_len - name_len)  # left justify

        desc = starter.get('description', '')
        desc = desc.splitlines()[0]  # keep only the first line
        end = wrap.fill(desc)

        # Replace the name with colored text
        line = " - ".join((start, end))

        # print the starter line to terminal
        print(line)


def print_starter_info(name, starter):
    """Print info for the starter with the given name.

    Parameters
    ----------
    name : str
        The project starter name.
    starter : dict
        A dict with entries on the project starter. These are the entries
        (values) from the generate_starters_dict function.
    """
    # Setup the textwrapper and cli settings
    name_clr = settings.cli_init_starter_name_color
    heading_clr = settings.cli_init_starter_subheadind_color

    wrap = TextWrapper(width=min(80, term_width()),
                       initial_indent='  ', subsequent_indent='  ')

    # Print the starter name and title
    if 'title' in starter:
        print("{} ({})".format(click.style(starter['title'], bold=True),
                               click.style(name, fg=name_clr)))
    else:
        print(click.style(name, fg=name_clr))

    # Print the description
    if 'description' in starter:
        description = dedent("  " + starter['description'])

        block = "\n"
        block += click.style("Description", fg=heading_clr) + '\n'
        block += wrap_with_newlines(description, wrapper=wrap)
        print(block)

    # Print the other fields
    fields = [field for field in starter.keys()
              if field not in ('title', 'description') and not
              field.startswith('_')]
    for field in fields:
        block = "\n"
        block += click.style(field.title(), fg=heading_clr) + '\n'
        block += wrap_with_newlines(starter[field], wrapper=wrap)
        print(block)

    # Print a list of files in the project starter
    filepaths = starter_filepaths(starter)
    if filepaths:
        block = "\n"
        block += click.style("Files", fg=heading_clr) + '\n'
        for filepath in sorted(filepaths.keys(), key=filepath_key):
            block += wrap.fill(str(filepath)) + '\n'
        print(block)


def is_empty(path):
    """Returns True if the path is a directory and it's empty."""
    path = pathlib.Path(path)
    if not path.is_dir():
        return False
    try:
        next(path.iterdir())
    except StopIteration:
        return True

    return False


def clone_starter(name, starter, out_dir):
    """Clone the given project starter to the out_dir.

    Parameters
    ----------
    name : str
        The project starter name.
    starter : dict
        A dict with entries on the project starter. These are the entries
        (values) from the generate_starters_dict function.
    out_dir : str
        The directory to create the starter template copy.
    """
    twidth = min(80, term_width())
    name_clr = settings.cli_init_starter_name_color

    out_dir = (pathlib.Path(out_dir) if out_dir is not None else
               pathlib.Path('.'))

    wrap = TextWrapper(width=twidth)

    # Check to see if the destination directory is empty
    msg = ("The directory '{}' is not empty; do you still want to initialize "
           "the project?")
    msg = wrap.fill(msg.format(out_dir))
    if not is_empty(out_dir) and not click.confirm(msg):
        exit()

    # Ok, we're good to clone the project starter
    filepaths = starter_filepaths(starter)
    max_filepath_len = max(len(filepath) for filepath in filepaths)

    for filepath in sorted(filepaths, key=filepath_key):
        full_filepath = filepaths[filepath]

        # Format the printed message
        src = filepath
        dest = out_dir / filepath

        start = "Copying {}{}".format(name + '/', src)
        start += " " * (max_filepath_len - len(str(src)) + 1)
        end = "-> {}".format(dest)

        # Colorize the text
        len_str = len(start + end)
        start = "Copying {}{}".format(click.style(name + '/', fg=name_clr),
                                      src)
        start += " " * (max_filepath_len - len(str(src)) + 1)

        if len_str > twidth:
            print("\n  ".join((start, end)))
        else:
            print("".join((start, end)))

        # Make the destination directory
        dest.parent.mkdir(parents=True, exist_ok=True)
        copy2(full_filepath, dest)



"""
Checkers for python packages.
"""
from .checker import Checker
from .utils import name_and_version
from .types import SoftwareDependency, SoftwareDependencyList
from ..software_dependencies import python_deps


class PythonChecker(Checker):
    """Checker for python dependencies.

    Attributes
    ----------
    freeze : Optional[Tuple[str]]
        If pip is available, this is populated with a tuple of strings for
        installed packages.
    """

    freeze = None

    def __init__(self, python_deps=python_deps):
        super().__init__('python', *python_deps.dependencies)

        # Setup pip freeze, if available
        try:
            # This is the correct location for pip > 10.0
            from pip._internal.operations import freeze
        except ImportError:
            try:
                from pip.operations import freeze
            except ImportError:
                freeze = None

        # Retrieve the the freeze list of strings, if available
        self.freeze = (tuple(i.lower() for i in freeze.freeze())
                       if freeze is not None else None)

    def check_packages_pip(self, packages=None):
        """Check the availability of specific python packages in pip.

        Parameters
        ----------
        listing : Union[List[str], Tuple[str]]
            Packages names with optional version specifier.
        """
        # if pip freeze isn't available, return None
        if self.freeze is None:
            return None

        # Get the packages
        if packages is not None:
            pass
        elif 'packages' in self:
            packages = self['packages']
        else:
            return None

        if isinstance(packages, SoftwareDependency):
            # Convert the listings to tuples of name, operator, version.
            # ex: 'python>=3.1' becomes 'python',
            #     <built-in function ge>, (3, 1)
            rv = name_and_version(packages.name)

            if rv is None:
                # If the name a version could not be parsed, return None and
                # do nothing more
                return None

            # Parse the package name, comparison operator (op) and the version
            # tuple
            name, op, ver = rv
            name = name.lower()

            # See if the package name is available in the freeze tuple
            matching_packages = [i for i in self.freeze if i.startswith(name)]

            # If there are no matching packages, return None and do nothing
            # more
            if len(matching_packages) == 0:
                return None

            # In this case, there is an installed package. See if the version
            # matches
            installed_package = matching_packages[0]
            rv = name_and_version(installed_package)

            if rv is None:
                # The installed package could not be parsed. Return None and
                # do nothing more
                return None

            installed_name, installed_op, installed_ver = rv

            # check the version
            if op is None and ver is None:
                # In this case, we don't care about the version--just that it's
                # installed. It is, so mark it as installed
                packages.available = True
            elif installed_ver is not None and op(installed_ver, ver):
                # Otherwise check the version number
                packages.available = True

        elif isinstance(packages, SoftwareDependencyList):
            # If the packages is a SoftwareDependencyList, process each package
            # individually
            for package in packages.dependencies:
                self.check_packages_pip(package)



"""
The Tornado main app
"""

import tornado.httpserver
import tornado.ioloop
import tornado.web

from .urls import url_patterns
from .handlers.server import server_template_path, ServerHandler
from .. import settings


class TornadoApp(tornado.web.Application):
    """The Tornado App for the built-in webserver"""
    def __init__(self, **kwargs):
        tornado.web.Application.__init__(self, url_patterns, **kwargs)


def get_app(in_path, out_dir, debug=False, **kwargs):
    """Create the Tornado app instance"""
    app = TornadoApp(in_path=in_path, out_dir=out_dir, debug=debug,
                     template_path=server_template_path,
                     default_handler_class=ServerHandler)
    return app


def run_server(in_path, out_dir, port=settings.default_port,
               debug=False):
    """Create and run the web-server.

    Parameters
    ----------
    in_path : str
        The filenames for project root documents.
    out_dir : str
        The target root directory to render files to.
    port : Optional[int]
        The network port to serve the web-server.
    debug : Optional[bool]
        If true, include debugging information.
    """
    app = get_app(in_path=in_path, out_dir=out_dir, debug=debug)

    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(port, '127.0.0.1')  # listen only to the localhost
    tornado.ioloop.IOLoop.instance().start()



"""
A CompositeBuilder for html files.
"""
from .target_builder import TargetBuilder
from .xhtml_builder import XHtmlBuilder
from ...paths import TargetPath
from ...utils.list import uniq
from ... import settings


class EpubBuilder(TargetBuilder):
    """A builder for epub files."""

    available = True
    priority = 1000
    infilepath_ext = '.dm'
    outfilepath_ext = '.epub'

    only_root = True
    add_parallel_builder = False
    add_render_builder = False

    _xhtml_builder = None
    _epub_builder = None
    _render_toc_context = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Populate the subbuilders
        self.create_subbuilders()

    def find_or_create_xhtml_builders(self, **kwargs):
        """Find or create the target XHtmlBuilder for all documents in the
        document tree."""
        # Get all documents in the document tree
        context = self.context
        env = self.env
        doc = context.document
        doc_list = doc.documents_list(only_subdocuments=False, recursive=True)

        xhtml_builders = []
        for doc in doc_list:
            builders = doc.context.setdefault('builders', dict())
            if '.xhtml' not in builders:
                xhtml_builder = XHtmlBuilder(env=env, context=doc.context,
                                             target='xhtml', **kwargs)

                builders['.xhtml'] = xhtml_builder
            else:
                xhtml_builder = builders['.xhtml']

            xhtml_builders.append(xhtml_builder)
        return xhtml_builders

    def xhtml_filepaths(self, xhtml_builders=None):
        """Return a flat list of xhtml, css, svg filepaths from the
        XHtmlBuilders"""
        # Find or create all of the xhtml builders for this document and
        # all subdocuments
        xhtml_builders = xhtml_builders or self.find_or_create_xhtml_builders()

        # Find all the xhtml filepaths for these xhtml_builders
        filepaths = [builder.outfilepath for builder in xhtml_builders]

        # Find all the xhtml filepaths for other dependencies
        for xhtml_builder in xhtml_builders:
            subbuilders = xhtml_builder._parallel_builder.subbuilders
            filepaths += [builder.outfilepath for builder in subbuilders]

        # Return filepaths for build xhtml files and other dependencies
        # created
        allowed_exts = set(settings.tracked_deps['.xhtml']) | {'.xhtml'}
        return uniq([filepath for filepath in filepaths
                     if filepath.suffix in allowed_exts])

    def has_toc_xhtml(self, xhtml_filepaths=None):
        """Evaluates whether there is a toc.xhtml file in the xhtml files."""
        xhtml_filepaths = xhtml_filepaths or self.xhtml_filepaths()
        return any(filepath.subpath == 'toc.xhtml'
                   for filepath in xhtml_filepaths)

    def create_toc_xhtml_builder(self, context=None,
                                 template_name='default/xhtml/toc.xhtml'):
        """Create a toc.xhtml file for the document.

        Parameters
        ----------
        context : :obj:`.document.DocumentContext`
            The document context for the document that owns this builder.
        outfilepath : Optional[:obj:`paths.TargetPath`]
            The outfilepath to use for the toc.xhtml
        template_name : Optional[str]
            The name of the template file to use in making the toc.
        """
        context = context or self.context
        cache_path = self.env.cache_path
        outfilepath = TargetPath(target_root=cache_path, target='xhtml',
                                 subpath='toc.xhtml')

        # Find the render builder class
        builder_cls = self.find_builder_cls(in_ext='.render')

        # Create a mock "toc" document for the toc_xhtml_builder.
        # This will require a new context (render_context), owned by this
        # builder so that the render builder can hold a weakref to it.
        render_context = context.filter(['toc', 'label_manager',
                                         'root_document', 'relative_links'])
        self._render_toc_context = render_context

        # Modify the new context and toc tag to mock a new toc document
        toc = render_context['toc'].copy()
        render_context['template'] = template_name
        render_context['toc'] = toc
        render_context['doc_id'] = 'toc.dm'

        toc.doc_id = 'toc.dm'
        toc.context = render_context

        # If no toc entries are present in the toc tag, modify the toc tag to
        # return entries to the documents in the document tree. This is because
        # epub docs require at least 1 reference
        if len(toc.get_labels()) == 0:
            toc.toc_kind = ('all', 'documents', 'expanded')

        builder = builder_cls(env=self.env, context=render_context,
                              outfilepath=outfilepath, render_ext='.xhtml',
                              use_media=False, use_cache=True, target='xhtml')
        return builder

    def create_subbuilders(self):
        """Create subbuilders needed by this builder"""
        # Get parameters needed to create builders
        env = self.env
        context = self.context

        # Reset this builder's subbuilders
        self.subbuilders.clear()

        # 1. Create the xhtml builders for all document files
        xhtml_builders = self.find_or_create_xhtml_builders()
        xhtml_filepaths = self.xhtml_filepaths(xhtml_builders=xhtml_builders)
        self.subbuilders += xhtml_builders

        # 2. Build the toc file, if needed
        if not self.has_toc_xhtml(xhtml_filepaths=xhtml_filepaths):
            toc_builder = self.create_toc_xhtml_builder(context=self.context)
            self.subbuilders.append(toc_builder)
            xhtml_filepaths.append(toc_builder.outfilepath)

        # 3. Create xhtml to epub builder
        xhtml2epub_cls = self.find_builder_cls(in_ext='.xhtml',
                                               out_ext='.epub')
        xhtml2epub = xhtml2epub_cls(env=env, parameters=xhtml_filepaths,
                                    context=context, target='epub',
                                    use_media=False, use_cache=True)
        self.subbuilders.append(xhtml2epub)

        # 4. Create a copy builder to be placed in the final epub directory
        copy_builder_cls = self.find_builder_cls(in_ext='.*', out_ext='.*')
        copy_builder = copy_builder_cls(env=env, target='epub',
                                        use_cache=False)
        self.subbuilders.append(copy_builder)

        # Setup the paths
        copy_builder.parameters = [xhtml2epub.outfilepath]
        copy_builder.outfilepath = self.outfilepath



"""
Tags for tables
"""
from .tag import Tag
from .data import Data
from .caption import Caption
from ..formats.tex import tex_env


class BaseTable(Tag):
    """A base tag for all tables"""

    active = False
    tex_env = 'table'
    html_name = 'table'
    html_class = None
    include_paragraphs = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Transfer the label id ('id') to the caption, if available. First,
        # find the caption tag, if available
        captions = [tag for tag in self.flatten(filter_tags=True)
                    if isinstance(tag, Caption)]

        for caption in captions:
            # Transfer the 'id' to the caption (but only the first)
            caption.label_id = self.attributes.pop('id', None)

            # Set the label kind for the caption as a table caption
            caption.kind = ('caption', 'table')

            # Make the caption a caption tag for html
            caption.html_name = 'caption'
            caption.html_class = 'tablecaption'

            # Create the label in the label_manager
            caption.create_label()

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        # Collect the contents of this table and prepare in a list
        content = content if content is not None else self.content
        content = [content] if not isinstance(content, list) else content

        # Render the html for the content
        content_tex = []
        for item in content:
            if item == self:
                # The flatten method will also return the self tag. To avoid
                # recursion, do not process this tag.
                continue
            elif isinstance(item, Data):
                # Process the data in the table into a tabular environment
                table_tex = item.tex_table(level=level + 1)
                col_types = 'l' * item.num_cols
                content_tex += tex_env('tabular', attributes=col_types,
                                       formatted_content=table_tex)
            elif isinstance(item, Tag):
                # Process other tags like captions
                content_tex.append(item.tex_fmt(level=level + 1))
            else:
                content_tex.append(item)

        # Return the formatted table
        return super().tex_fmt(content=content_tex, attributes=attributes,
                               level=level, **kwargs)

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1, **kwargs):
        # Setup the html class, if specified
        attributes = (attributes if attributes is not None else
                      self.attributes.filter(target='.html'))
        if self.html_class is not None:
            if 'class' in attributes:
                attributes['class'] += " " + self.html_class
            else:
                attributes['class'] = self.html_class

        # Collect the contents of this table and prepare in a list
        content = content if content is not None else self.content
        content = [content] if not isinstance(content, list) else content

        # Render the html for the content
        content_html = []
        for item in content:
            if item == self:
                # The flatten method will also return the self tag. To avoid
                # recursion, do not process this tag.
                continue
            elif isinstance(item, Data):
                # Process the data in the table
                content_html += item.html_table(format_func=format_func,
                                                method=method, level=level + 1)
            elif isinstance(item, Tag):
                # Process other tags like captions. Use either 'html_fmt' or
                # 'xhtml_fmt' functions, as indicated by the format_func
                # variable
                html = getattr(item, format_func)(method=method,
                                                  format_func=format_func,
                                                  level=level + 1)
                content_html.append(html)
            else:
                content_html.append(item)

        # Return the formatted table
        return super().html_fmt(content=content_html, attributes=attributes,
                                method=method, level=level, **kwargs)

    def xhtml_fmt(self, format_func='xhtml_fmt', method='xhtml', **kwargs):
        return self.html_fmt(format_func=format_func, method=method, **kwargs)


class MarginTable(BaseTable):
    """The @margintable tag"""

    aliases = ('margintbl',)
    html_class = 'margintbl'
    tex_env = 'margintable'
    active = True


class Table(BaseTable):
    """The @table tag"""

    aliases = ('tbl',)
    html_class = 'table'
    tex_env = 'table'
    active = True


class FullTable(BaseTable):
    """The @fulltable tag"""

    aliases = ('ftbl', 'fulltbl')
    html_class = 'fulltable'
    tex_env = 'table*'
    active = True



"""
Utilities for builders and environments
"""
import pathlib

from .deciders.utils_hash import hash_items
from ..paths import SourcePath, TargetPath
from ..paths.utils import rename


def sort_key(parameter):
    """Sort key function for a series of parameters to give a consisting
    ordering."""
    if (isinstance(parameter, tuple) or
       isinstance(parameter, list) and parameter):
        return str(parameter[0])
    else:
        return str(parameter)


def generate_mock_parameters(env, parameters, project_root=None, subpath=None,
                             ext=None, context=None, gen_hash=True):
    """Generate a mock set of parameters.

    This function is used by builders, like JinjaRender and SaveTempFile, that
    do not use an input file but need to save to an output file. It includes
    an option to rename the filename with a hash.

    Parameters
    ----------
    env : :obj:`.builders.Environment`
        The build environment.
    parameters : Tuple[:obj:`.paths.SourcePath`, str, tuple, list]
        The parameters for the builder.
    project_root : Optional[Union[:obj:`paths.SourcePath`, str]]
        If given, use the specified project root for the mock infilepath.
    subpath : Optional[Union[:obj:`paths.SourcePath`, str]]
        If given, use the specified subpath for the mock infilepath.
    ext : Optional[str]
        If given, use the specified extension for the mock infilepath.
    gen_hash : Optional[bool]
        If True, append the hash from the given parameters to append to the
        infilepath filename.
    """
    # Setup the parameters
    parameters = parameters or []
    parameters = (parameters if isinstance(parameters, list) or
                  isinstance(parameters, tuple) else [parameters])
    subpath = pathlib.Path(subpath) if subpath is not None else None

    # First get the project root
    project_root = (project_root or env.project_root)

    # Next get the subpath
    if subpath is None:
        filepaths = [i for i in parameters if hasattr(i, 'subpath')]

        if filepaths:
            # First check to see if any of the passed parameters have a subpath
            subpath = filepaths[0].subpath.with_suffix('')
        elif context is not None and 'src_filepath' in context:
            # Otherwise construct the filepath from the context's src_filepath
            src_filepath = context['src_filepath']
            subpath = src_filepath.subpath.with_suffix('')

    # Split the subpath and filename, if available and these aren't
    # specified.
    filename = getattr(subpath, 'name', None)
    subpath = getattr(subpath, 'parent', None)

    # Generate a hash
    if gen_hash:
        # The hash is constructed from the parameters.
        hash_value = hash_items(*parameters, sort=True)[:12]  # truncate hash
    else:
        hash_value = None

    # Generate a filename with the hash
    filename = "_".join(filter(bool, [filename, hash_value]))  # Remove None
    filename = pathlib.Path(filename)

    # Set the extension, if specified
    if ext:
        filename = filename.with_suffix(ext)

    if subpath:
        return SourcePath(project_root=project_root,
                          subpath=subpath / filename)
    else:
        return SourcePath(project_root=project_root, subpath=filename)


def generate_outfilepath(env, parameters, target=None, append=None, ext=None,
                         use_cache=False, use_media=False):
    """Given a set of parameters, generate an outfilepath.

    Parameters
    ----------
    env : :obj:`.builders.Environment`
        The build environment.
    parameters : Tuple[:obj:`pathlib.Path`]
        The parameters for the builder. The first pathlib.Path will be used.
    target : Optional[str]
        If specified, use the given target as a subdirectory in the
        target_root.
    append : Optional[str]
        If specified, append the given string to the returned filename
    ext : Optional[str]
        If specified, return a cache_filepath with the given target format.
    use_cache : bool
        If True, return a path in the cache directory.
    use_media : bool
        If True, return a path with a subpath prepended with the media_path.
        ex: 'media/'

    Returns
    -------
    outfilepath : Union[:obj:`.paths.TargetPath`, None]
        The outfilepath based on the parameters given, or None if no
        outfilepath could be generated

    .. note :: When dealing with absolute paths, only the filename is kept
               in formulating the subpath. So for

               target_root = '.'
               target = 'html'
               use_media = True
               parameters = ['/usr/local/bin/2to3']

               The generated outfilepath will be:

               TargetPath('html/media/2to3')
    """
    # Find the first valid infilepath that is not an absolute path
    parameters = (parameters if isinstance(parameters, list) or
                  isinstance(parameters, tuple) else [parameters])
    parameters = [fp for fp in parameters if isinstance(fp, pathlib.Path)]

    if len(parameters) == 0:
        return None

    # Find the first infilepath that isn't an absolute path
    infilepath = parameters[0]

    # Formulate the target_root
    target_root = env.cache_path if use_cache else env.target_root
    media_path = env.media_path if use_media else None

    # Formulate the target
    if isinstance(target, str):
        target = target.strip('.')
    elif hasattr(infilepath, 'target'):
        target = infilepath.target
    else:
        target = None

    # Formulate the subpath
    subpath = (infilepath.subpath if hasattr(infilepath, 'subpath') else
               infilepath)

    # Convert the subpath to a filename, if it's an absolute path
    if hasattr(subpath, 'is_absolute') and subpath.is_absolute():
        subpath = pathlib.Path(subpath.name)

    # Add media_path, if specified
    if media_path and not str(subpath).startswith(media_path):
        subpath = rename(path=media_path / subpath, append=append,
                         extension=ext)
    else:
        subpath = rename(path=subpath, append=append, extension=ext)

    # Return the new TargetPath
    return (TargetPath(target_root=target_root, target=target, subpath=subpath)
            if target else
            TargetPath(target_root=target_root, subpath=subpath))



"""
A build environment to manage and execute builders.
"""
import pathlib
from itertools import chain

from .deciders import Decider
from .scanners import Scanner
from .composite_builders import ParallelBuilder
from ..document import Document
from ..paths import SourcePath, TargetPath
from .. import settings


class Environment(object):
    """A project environment for rendering documents using builders.

    The build project environment has the following tasks:
        1. Setup the project_root and target_root
        2. Setup the default decider for builders
        3. Setup the default scanner for builders
        4. Setup the root document.

    Parameters
    ----------
    src_filepath : Union[str, :obj:`pathlib.Path`]
        The path for the disseminate source file of the root document.
    target_root : Optional[Union[str, :obj:`pathlib.Path`]]
        The (optional) path for the output root directory.
    parent_context : Optional[:obj:`.context.BaseContext`]
        The document context to use as the parent_context for the root
        document. Typically, the default_context from the settings is used as
        the parent context.
    """

    decider = None
    scanner = None

    root_document = None
    context = None
    project_root = None
    target_root = None

    _cache_path = None
    _concrete_builders = None

    def __init__(self, src_filepath, target_root=None, parent_context=None):

        # Setup the decider
        decider_cls = [cls for cls in Decider.__subclasses__()
                       if cls.__name__ == settings.default_decider]
        assert decider_cls, ("The decider class '{}' does not "
                             "exist".format(settings.default_decider))
        decider_cls = decider_cls[0]
        self.decider = decider_cls(env=self)

        # Setup the scanner
        self.scanner = Scanner

        # Setup the paths for the parent_context
        if not isinstance(src_filepath, SourcePath):
            # Make sure the src_filepath is a Source path
            src_filepath = pathlib.Path(src_filepath)
            src_filepath = SourcePath(project_root=src_filepath.parent,
                                      subpath=src_filepath.name)
        project_root = SourcePath(project_root=src_filepath.project_root)
        self.project_root = project_root

        if target_root is not None:
            # Make sure the target_root is a TargetPath
            target_root = TargetPath(target_root=target_root)
        else:
            target_root = self.get_target_root(project_root)
        self.target_root = target_root

        # Setup the root document
        root_document = Document(src_filepath=src_filepath,
                                 parent_context=parent_context,
                                 environment=self)
        self.context = root_document.context
        self.root_document = root_document

    def __repr__(self):
        return "Environment({})".format(self.project_root)

    @property
    def cache_path(self):
        """The path to the directory for storing cached files."""
        if self._cache_path is None:
            cache_path = SourcePath(project_root=self.target_root,
                                    subpath=settings.cache_path)
            cache_path.mkdir(parents=True, exist_ok=True)
            self._cache_path = cache_path
        return self._cache_path

    @property
    def media_path(self):
        """The path for to prepend to subpaths for media files."""
        if self.context:
            return self.context.get('media_path', None)
        return None

    @property
    def name(self):
        """The name of the environment"""
        root_doc = self.root_document
        src_filepath = (root_doc.src_filepath.subpath if root_doc is not None
                        else None)
        return str(src_filepath) if src_filepath is not None else ''

    @staticmethod
    def get_target_root(project_root):
        """Determine the target_root directory from the project_root."""
        if project_root.match(settings.document_src_directory):
            # If the project_root is in a src directory, use the directory
            # above this directory
            return TargetPath(target_root=project_root.parent)
        else:
            # Otherwise just use the same directory as the src directory
            return TargetPath(target_root=project_root)

    @staticmethod
    def _get_src_filepaths(root_path='',
                           document_extension=settings.document_extension):
        """Find the root src_filepaths from a given root_path"""
        # Create a glob pattern to get all of the disseminate files in the path
        # and remove the filenames to retain the unique paths. Convert the
        # paths to strings so that they can be sorted more easily.
        glob_pattern = pathlib.Path('**', '*' + document_extension)
        paths = {p.parent for p in root_path.glob(str(glob_pattern))}

        # Only keep entries that are basepaths
        basepaths = set()

        for path in paths:
            # See if its a base path based on whether any of its parent is
            # already located in the paths
            parents = path.parents
            if any(parent in paths for parent in parents):
                continue

            # The basepath is unique; add it to basepaths
            basepaths.add(path)

        # Generate the unique root paths as pathlib.Path objects.
        project_paths = [SourcePath(project_root=basepath)
                         for basepath in basepaths]

        src_filepaths = []

        for project_path in project_paths:
            # Find the src_filepaths for the root documents (non-recursive)
            glob_pattern = pathlib.Path('*' + document_extension)
            glob = list(project_path.glob(str(glob_pattern)))
            subpaths = [filepath.relative_to(project_path)
                        for filepath in glob]
            src_filepaths += [SourcePath(project_root=project_path,
                                         subpath=subpath)
                              for subpath in subpaths]
        return src_filepaths

    @staticmethod
    def create_environments(root_path='', target_root=None,
                            document_extension=settings.document_extension):
        """Create environments from root documents found in the given
        root_path.

        Parameters
        ----------
        root_path : Optional[Union[str, :obj:`pathlib.Path`]]
            The path to search for root documents. By default, it is the
            current directory.
        target_root : Optional[Union[str, :obj:`pathlib.Path`]]
            The (optional) path for the output root directory.
        document_extension : Optional[str]
            The file extension for disseminate documents. ex: '.dm'

        Returns
        -------
        environments : List[:obj:`.builders.Environment`]
        """
        # expand the user for the subpath directory
        path = pathlib.Path(root_path).expanduser()

        # Get the src_filepaths for the root documents
        if path.is_file():
            # src_filepaths from the file directly
            src_filepaths = [SourcePath(project_root=path.parent,
                                        subpath=path.name)]
        else:
            # or through the directory with glob patterns (_get_src_filepaths)
            get_fps = Environment._get_src_filepaths
            fps = get_fps(root_path=path,
                          document_extension=document_extension)
            src_filepaths = fps

        return [Environment(src_filepath=src_filepath, target_root=target_root)
                for src_filepath in src_filepaths]

    def create_root_builder(self, document=None):
        """Create a root parallel builder for the target builders of the given
        document or the root document.

        Parameters
        ----------
        document : Optional[:obj:`.document.Document`]
            The document to create a root builder for.

        Returns
        -------
        root_builder : :obj:`.composite_builders.ParallelBuilder`
            The root (parallel) builder.
        """
        root_builder = ParallelBuilder(env=self)
        root_builder.clear_done = True

        # Get all of the documents
        subbuilders = self.collect_target_builders(document=document)
        root_builder.subbuilders += subbuilders

        return root_builder

    def collect_target_builders(self, document=None):
        """Return all target builders for the root_document and all
        sub-documents.

        Parameters
        ----------
        document : Optional[:obj:`.document.Document`]
            The document to create a root builder for.

        Returns
        -------
        target_builders : List[:obj:`.builders.Builder`
            The list of target builders
        """
        document = document if document is not None else self.root_document
        documents = document.documents_list(only_subdocuments=False,
                                            recursive=True)
        return chain(*[doc.context['builders'].values()
                       if 'builders' in doc.context else []
                       for doc in documents])

    def build(self, complete=True):
        """Run the build.

        Parameters
        ----------
        complete : Optional[bool]
            If True, run the build until it has completed
            If False, start the build in the background.

        Returns
        -------
        status : str
            The current status of the build.
        """
        root_builder = self.create_root_builder()
        return root_builder.build(complete=complete)
